"""
Read currency exchange rates.
This functionality could be used to display the exchange rate graph, for example.

The first (and only) parameter is the name of the GnuCash file to use. If not set,
'test.gnucash' is used.
"""
# pylint: disable=invalid-name
import sys
import piecash
from piecash import Commodity

# Variables
filename = sys.argv[1]
if filename is None:
    print("You need to specify a valid .gnucash file to use.")
    filename = "test.gnucash"

symbol = "AUD"
####################################

with piecash.open_book(filename, open_if_lock=True) as book:
    # , readonly=False,

    # Get all commodities.
    # The commodities (including currencies) in the book are only those used in accounts.
    #commodities = book.commodities

    # Get all the currencies in the book (i.e. for update).
    print("All currencies used in the book:")
    currencies = book.currencies
    for c in currencies:
        print(c)

    # Accessing individual records.

    print("\nSelected single currency details (" + symbol + "):")
    cdty = book.get(Commodity, namespace="CURRENCY", mnemonic=symbol)

    # Accessing attributes of a commodity.
    print("Commodity namespace={cdty.namespace}\n"
        "          mnemonic={cdty.mnemonic}\n"
        "          cusip={cdty.cusip}\n"
        "          fraction={cdty.fraction}".format(cdty=cdty))

    # Loop through the existing commodity prices.
    # This can be used to fetch the points for a price graph.
    print("\nHistorical prices:")
    for pr in cdty.prices:
        print("Price date={pr.date}"
            "      value={pr.value} {pr.currency.mnemonic}/{pr.commodity.mnemonic}".format(pr=pr))

    # List of accounts which use the commodity:
    #cdty.accounts



import csv
from pathlib import Path

from piecash import open_book, Account

GNUCASH_BOOK = "../gnucash_books/simple_sample.gnucash"

# open the book and the export file
with open_book(GNUCASH_BOOK, readonly=True, open_if_lock=True) as book:
    # show accounts
    print(book.accounts)
    print("Number of splits in the book:", len(book.splits))
    # select the 3rd account
    account = book.accounts[2]
    print(account, " has splits: ", account.splits)

    # delete the account from the book
    book.delete(account)
    # flush the change
    book.flush()
    # check the account has disappeared from the book and its related split too
    print(book.accounts)
    print("Number of splits in the book:", len(book.splits))

    # even if the account object and its related object still exists
    print(account, " has splits: ", account.splits)

    # do not forget to save the book if you want
    # your changes to be saved in the database



import datetime
import decimal
import sys
import uuid
from enum import Enum
from importlib import import_module

from sqlalchemy import Column, VARCHAR, INTEGER, REAL, BIGINT, types, event, Index
from sqlalchemy.orm import relation, foreign, object_session, backref

from ._common import CallableList
from ._common import hybrid_property_gncnumeric
from .sa_extra import _DateTime, DeclarativeBase, _Date



class KVP_Type(Enum):
    KVP_TYPE_INVALID = -1
    KVP_TYPE_GINT64 = 1
    KVP_TYPE_DOUBLE = 2
    KVP_TYPE_NUMERIC = 3
    KVP_TYPE_STRING = 4
    KVP_TYPE_GUID = 5
    KVP_TYPE_TIMESPEC = 6
    KVP_TYPE_BINARY = 7
    KVP_TYPE_GLIST = 8
    KVP_TYPE_FRAME = 9
    KVP_TYPE_GDATE = 10


pytype_KVPtype = {
    int: KVP_Type.KVP_TYPE_GINT64,
    float: KVP_Type.KVP_TYPE_DOUBLE,
    decimal.Decimal: KVP_Type.KVP_TYPE_NUMERIC,
    dict: KVP_Type.KVP_TYPE_FRAME,
    list: KVP_Type.KVP_TYPE_GLIST,
    # to fill
}

KVPtype_fields = {
    KVP_Type.KVP_TYPE_GINT64: 'int64_val',
    KVP_Type.KVP_TYPE_DOUBLE: 'double_val',
    KVP_Type.KVP_TYPE_STRING: 'string_val',
    KVP_Type.KVP_TYPE_GUID: 'guid_val',
    KVP_Type.KVP_TYPE_TIMESPEC: 'timespec_val',
    KVP_Type.KVP_TYPE_GDATE: 'gdate_val',
    KVP_Type.KVP_TYPE_NUMERIC: ('numeric_val_num', 'numeric_val_denom'),
    KVP_Type.KVP_TYPE_FRAME: 'guid',
    KVP_Type.KVP_TYPE_GLIST: 'guid',
}


class SlotType(types.TypeDecorator):
    """Used to customise the DateTime type for sqlite (ie without the separators as in gnucash
    """
    impl = INTEGER

    def process_bind_param(self, value, dialect):
        if value is not None:
            return value.value

    def process_result_value(self, value, dialect):
        if value is not None:
            return KVP_Type(value)


class DictWrapper(object):
    def __contains__(self, key):
        for sl in self.slots:
            if sl.name == key:
                return True
        else:
            return False

    def __getitem__(self, key):
        assert not isinstance(key, int), \
            "You are accessing slots with an integer (={}) while a string is expected".format(key)
        keys = key.split("/", 1)
        key = keys[0]
        for sl in self.slots:
            if sl.name == key:
                break
        else:
            raise KeyError("No slot exists with name '{}'".format(key))
        if len(keys) > 1:
            return sl[keys[1]]
        else:
            return sl  # .value

    def __setitem__(self, key, value):
        keys = key.split("/", 1)
        key = keys[0]
        for sl in self.slots:
            if sl.name == key:
                break
        else:
            # new key
            if len(keys) > 1:
                if isinstance(self, SlotFrame):
                    sf = SlotFrame(name=self._name + "/" + key,
                                   obj_guid=self.guid_val)
                else:
                    sf = SlotFrame(name=key,
                                   obj_guid=self.guid)
                sf[keys[1]] = value
                self.slots.append(sf)
            else:
                self.slots.append(slot(parent=self, name=key, value=value))

            return

        if len(keys) > 1:
            sl[keys[1]] = value
            return
        # assign if type is correct
        if isinstance(value, sl._python_type):
            sl.value = value
        else:
            raise TypeError("Type of '{}' is not one of {}".format(value, sl._python_type))

    def __delitem__(self, key):
        if isinstance(key, slice):
            # delete all
            del self.slots[key]
            return
        keys = key.split("/", 1)
        for i, sl in enumerate(self.slots):
            if sl.name == keys[0]:
                break
        else:
            raise KeyError("No slot exists with name '{}'".format(key))
        if len(keys) > 1:
            del sl[keys[1]]
        else:
            del self.slots[i]

    def iteritems(self):
        for sl in self.slots:
            yield sl.name, sl

    def get(self, key, default=None):
        try:
            return self[key].value
        except KeyError:
            return default


class Slot(DeclarativeBase):
    __tablename__ = 'slots'

    __table_args__ = (
        Index('slots_guid_index', 'obj_guid'),
        {'sqlite_autoincrement': True, }
    )

    # column definitions
    id = Column('id', INTEGER(), primary_key=True, nullable=False, autoincrement=True)
    obj_guid = Column('obj_guid', VARCHAR(length=32), nullable=False)
    _name = Column('name', VARCHAR(length=4096), nullable=False)

    @property
    def name(self):
        if self._name:
            return self._name.split("/")[-1]
        else:
            return self._name

    @name.setter
    def name(self, value):
        self._name = value

    slot_type = Column('slot_type', SlotType(), nullable=False)

    __mapper_args__ = {
        'polymorphic_on': slot_type,
    }

    def __init__(self, name, value=None, obj_guid=None):
        self.name = name
        if value is not None:
            self.value = value
        if obj_guid is not None:
            self.obj_guid = obj_guid

    def __str__(self):
        return "<{} {}={!r}>".format(self.__class__.__name__, self.name, self.value)


class SlotSimple(Slot):
    __mapper_args__ = {
        'polymorphic_identity': -1,
    }

    _python_type = ()

    @property
    def value(self):
        return getattr(self, self._field)

    @value.setter
    def value(self, value):
        setattr(self, self._field, value)

    def __eq__(self, other):
        return (isinstance(other, self.__class__)
                and self.name == other.name
                and self.value == other.value
                )


def define_simpleslot(postfix, pytype, KVPtype, field, col_type, col_default):
    cls = type(
        'Slot{}'.format(postfix),
        (SlotSimple,),
        {
            "__mapper_args__": {'polymorphic_identity': KVPtype},
            field: Column(field, col_type, default=col_default),
            "_field": field,
            "_python_type": pytype,
        }
    )
    return cls


SlotInt = define_simpleslot(postfix="Int",
                            pytype=(int,),
                            KVPtype=KVP_Type.KVP_TYPE_GINT64,
                            field="int64_val",
                            col_type=BIGINT(),
                            col_default=0,
                            )
SlotString = define_simpleslot(postfix="String",
                               pytype=(str,),
                               KVPtype=KVP_Type.KVP_TYPE_STRING,
                               field="string_val",
                               col_type=VARCHAR(length=4096),
                               col_default=None,
                               )
SlotDouble = define_simpleslot(postfix="Double",
                               pytype=(float,),
                               KVPtype=KVP_Type.KVP_TYPE_DOUBLE,
                               field="double_val",
                               col_type=REAL(),
                               col_default=0,
                               )
SlotTime = define_simpleslot(postfix="Time",
                             pytype=(datetime.time,),
                             KVPtype=KVP_Type.KVP_TYPE_TIMESPEC,
                             field="timespec_val",
                             col_type=_DateTime(),
                             col_default=None,
                             )


class SlotFrame(DictWrapper, Slot):
    __mapper_args__ = {
        'polymorphic_identity': KVP_Type.KVP_TYPE_FRAME
    }
    _python_type = (dict,)

    guid_val = Column('guid_val', VARCHAR(length=32))

    slots = relation('Slot',
                     primaryjoin=foreign(Slot.obj_guid) == guid_val,
                     cascade='all, delete-orphan',
                     collection_class=CallableList,
                     single_parent=True,
                     backref=backref("parent", remote_side=guid_val),

                     )

    @property
    def value(self):
        # convert to dict
        return {sl.name: sl.value for sl in self.slots}

    @value.setter
    def value(self, value):
        self.slots = [slot(parent=self, name=k, value=v) for k, v in value.items()]

    def __init__(self, **kwargs):
        self.guid_val = uuid.uuid4().hex
        super(SlotFrame, self).__init__(**kwargs)


class SlotList(SlotFrame):
    __mapper_args__ = {
        'polymorphic_identity': KVP_Type.KVP_TYPE_GLIST
    }
    _python_type = (list,)

    @property
    def value(self):
        # convert to dict
        return [sl.value for sl in self.slots]

    @value.setter
    def value(self, value):
        self.slots = [slot(parent=self, name=str(i), value=v) for i, v in enumerate(value)]

    def __init__(self, **kwargs):
        self.guid_val = uuid.uuid4().hex
        super(SlotFrame, self).__init__(**kwargs)


@event.listens_for(SlotFrame.slots, 'remove')
def remove_slot(target, value, initiator):
    s = object_session(value)
    if value in s.new:
        s.expunge(value)
    else:
        s.delete(value)


class SlotGUID(SlotFrame):
    __mapper_args__ = {
        'polymorphic_identity': KVP_Type.KVP_TYPE_GUID
    }
    _python_type = (DeclarativeBase,)

    # add
    _mapping_name_class = {
        'from-sched-xaction': 'piecash.core.transaction.ScheduledTransaction',
        'account': 'piecash.core.account.Account',
        'invoice-guid': 'piecash.business.invoice.Invoice',
        'peer_guid': 'piecash.core.transaction.Split',
        'gains-split': 'piecash.core.transaction.Split',
        'gains-source': 'piecash.core.transaction.Split',
        'default-currency': 'piecash.core.commodity.Commodity',
    }

    @property
    def Class(self):
        name, guid = self.name, self.guid_val
        if name.startswith('CURRENCY::'):
            # handle capital gain account
            class_to_retrieve = 'piecash.core.account.Account'
        else:
            class_to_retrieve = self._mapping_name_class.get(name, None)
            if class_to_retrieve is None:
                raise ValueError(
                    "Smart retrieval of GUID slot with name '{}' is not yet supported. "
                    "Need to retrieve proper object type in kvp module "
                    "(add in SlotGUID._mapping_name_class)".format(name))
        class_module, class_name = class_to_retrieve.rsplit('.', 1)
        mod = import_module(class_module)
        Class = getattr(mod, class_name)
        return Class

    @property
    def value(self):
        return object_session(self).query(self.Class).filter_by(guid=self.guid_val).one()

    @value.setter
    def value(self, value):
        assert isinstance(value, self.Class)
        self.guid_val = value.guid


def get_all_subclasses(cls):
    all_subclasses = []

    direct_subclasses = cls.__subclasses__()

    all_subclasses.extend(direct_subclasses)

    for subclass in direct_subclasses:
        all_subclasses.extend(get_all_subclasses(subclass))

    return all_subclasses


def slot(parent, name, value):
    if isinstance(parent, SlotFrame):
        name = parent._name + "/" + name
        guid_parent = parent.guid_val
    else:
        guid_parent = parent.guid

    # handle datetime before others (as otherwise can be mixed with date)
    if isinstance(value, datetime.datetime):
        return SlotTime(name=name, value=value, obj_guid=guid_parent)

    for cls in get_all_subclasses(Slot):
        if isinstance(value, cls._python_type) and cls != SlotFrame and cls != SlotList:
            return cls(name=name, value=value, obj_guid=guid_parent)

    if isinstance(value, dict):
        # transform a dict to Frame/Slots
        sf = SlotFrame(name=name, obj_guid=guid_parent)
        for k, v in value.items():
            sl = slot(parent=sf, name=k, value=v)
            sl.parent = sf
        return sf

    if isinstance(value, list):
        # transform a list to List/Slots
        sf = SlotList(name=name)
        for i, v in enumerate(value):
            sl = slot(parent=sf, name=str(i), value=v)
            sl.parent = sf
        return sf

    raise ValueError("Cannot handle type of '{}'".format(value))


class SlotNumeric(Slot):
    __mapper_args__ = {
        'polymorphic_identity': KVP_Type.KVP_TYPE_NUMERIC
    }
    _python_type = (tuple, decimal.Decimal)

    _numeric_val_num = Column('numeric_val_num', BIGINT(), nullable=True, default=0)
    _numeric_val_denom = Column('numeric_val_denom', BIGINT(), nullable=True, default=1)
    value = hybrid_property_gncnumeric(_numeric_val_num, _numeric_val_denom)


SlotDate = define_simpleslot(postfix="Date",
                             pytype=(datetime.date,),
                             KVPtype=KVP_Type.KVP_TYPE_GDATE,
                             field="gdate_val",
                             col_type=_Date(),
                             col_default=None,
                             )



import datetime
import os
import shutil
import socket
from collections import defaultdict

from sqlalchemy import event, Column, VARCHAR, INTEGER, Table, PrimaryKeyConstraint
from sqlalchemy.sql.ddl import DropConstraint, DropIndex
from sqlalchemy_utils import database_exists

from piecash.core import factories
from .book import Book
from .._common import GnucashException
from ..sa_extra import create_piecash_engine, DeclarativeBase, Session

# version of tables changed between 2.6 and 3.0
#   ('invoices', 4)
#   ('prices', 3)
#   ('slots', 4)
#   ('transactions', 4)
#   ('entries', 4)

version_supported = {
    '2.6': {
        'Gnucash': 2062100,
        'Gnucash-Resave': 19920,
        'accounts': 1,
        'billterms': 2,
        'books': 1,
        'budget_amounts': 1,
        'budgets': 1,
        'commodities': 1,
        'customers': 2,
        'employees': 2,
        'entries': 3,
        'invoices': 3,
        'jobs': 1,
        'lots': 2,
        'orders': 1,
        'prices': 2,
        'recurrences': 2,
        'schedxactions': 1,
        'slots': 3,
        'splits': 4,
        'taxtable_entries': 3,
        'taxtables': 2,
        'transactions': 3,
        'vendors': 1,
    },
    '3.0': {
        'Gnucash': 3000000,
        'Gnucash-Resave': 19920,
        'accounts': 1,
        'billterms': 2,
        'books': 1,
        'budget_amounts': 1,
        'budgets': 1,
        'commodities': 1,
        'customers': 2,
        'employees': 2,
        'entries': 4,
        'invoices': 4,
        'jobs': 1,
        'lots': 2,
        'orders': 1,
        'prices': 3,
        'recurrences': 2,
        'schedxactions': 1,
        'slots': 4,
        'splits': 4,
        'taxtable_entries': 3,
        'taxtables': 2,
        'transactions': 4,
        'vendors': 1,
    },
    '3.7': {
        'Gnucash': 3000001,
        'Gnucash-Resave': 19920,
        'accounts': 1,
        'billterms': 2,
        'books': 1,
        'budget_amounts': 1,
        'budgets': 1,
        'commodities': 1,
        'customers': 2,
        'employees': 2,
        'entries': 4,
        'invoices': 4,
        'jobs': 1,
        'lots': 2,
        'orders': 1,
        'prices': 3,
        'recurrences': 2,
        'schedxactions': 1,
        'slots': 4,
        'splits': 5,
        'taxtable_entries': 3,
        'taxtables': 2,
        'transactions': 4,
        'vendors': 1,
    },
    '4.1': {
        'Gnucash': 4000001,
        'Gnucash-Resave': 19920,
        'accounts': 1,
        'billterms': 2,
        'books': 1,
        'budget_amounts': 1,
        'budgets': 1,
        'commodities': 1,
        'customers': 2,
        'employees': 2,
        'entries': 4,
        'invoices': 4,
        'jobs': 1,
        'lots': 2,
        'orders': 1,
        'prices': 3,
        'recurrences': 2,
        'schedxactions': 1,
        'slots': 4,
        'splits': 5,
        'taxtable_entries': 3,
        'taxtables': 2,
        'transactions': 4,
        'vendors': 1,
    }
}

# this is not a declarative as it is used before binding the session to an engine.
gnclock = Table(u'gnclock', DeclarativeBase.metadata,
                Column('hostname', VARCHAR(length=255)),
                Column('pid', INTEGER()),
                )


class Version(DeclarativeBase):
    """The declarative class for the 'versions' table.
    """
    __tablename__ = 'versions'

    __table_args__ = {}

    # column definitions
    # : The name of the table
    table_name = Column('table_name', VARCHAR(length=50), primary_key=True, nullable=False)
    #: The version for the table
    table_version = Column('table_version', INTEGER(), nullable=False)

    def __init__(self, table_name, table_version):
        self.table_name = table_name
        self.table_version = table_version

    def __str__(self):
        return "Version<{}={}>".format(self.table_name, self.table_version)


def build_uri(sqlite_file=None,
              uri_conn=None,
              db_type=None,
              db_user=None,
              db_password=None,
              db_name=None,
              db_host=None,
              db_port=None,
              check_same_thread=True,
              ):
    """Create the connection string in function of some choices.

    :param str sqlite_file: a path to an sqlite3 file (only used if uri_conn is None)
    :param str uri_conn: a sqlalchemy connection string
    :param str db_type: type of database in ["postgres","mysql"]
    :param str db_user: username of database
    :param str db_password: password for the use of database
    :param str db_name: name of database
    :param str db_host: host of database
    :param int db_port: port of database
    :param bool check_same_thread: sqlite flag that restricts connection use to the thread that created (see False for use in ipython/flask/... but read first https://docs.python.org/3/library/sqlite3.html)

    :return: the connection string
    :rtype: str
    """
    db_config = (db_type, db_host, db_port, db_name, db_user, db_password)
    db_config_isdefined = map(lambda x: x is not None, db_config[:-1])
    if any(db_config_isdefined):
        if not all(db_config_isdefined):
            raise ValueError("When using db_* arguments, all must be specified : {}".format(db_config))

        uri_conn = {"postgres": "postgresql://{username}:{password}@{host}:{port}/{name}",
                    "mysql": "mysql+pymysql://{username}:{password}@{host}:{port}/{name}?charset=utf8",
                    }[db_type].format(username=db_user, password=db_password,
                                      host=db_host, port=db_port,
                                      name=db_name)

    # db_postgres_uri = "postgresql://postgres:{pwd}@localhost:5432/foo".format(pwd=pg_password)
    # db_mysql_uri = "mysql+pymysql://travis:@localhost/foo?charset=utf8"
    # db_sqlite_uri = "sqlite:///{}".format(db_sqlite)

    if sqlite_file and uri_conn:
        raise ValueError("Only one of 'sqlite_file' or 'uri_conn' argument can be defined")

    if uri_conn is None:
        # fallback on sqlite
        if sqlite_file:
            if isinstance(sqlite_file, str) and sqlite_file.startswith("sqlite:///"):
                # already have the protocol specified.
                uri_conn = sqlite_file
            else:
                uri_conn = "sqlite:///{}".format(sqlite_file)
        else:
            uri_conn = "sqlite:///:memory:"
        if check_same_thread is False:
            uri_conn = uri_conn + "?check_same_thread=False"

    return uri_conn


def create_book(sqlite_file=None,
                uri_conn=None,
                currency="EUR",
                overwrite=False,
                keep_foreign_keys=False,
                db_type=None,
                db_user=None,
                db_password=None,
                db_name=None,
                db_host=None,
                db_port=None,
                check_same_thread=True,
                pg_template="template0",
                **kwargs):
    """Create a new empty GnuCash book. If both sqlite_file and uri_conn are None, then an "in memory" sqlite book is created.

    :param str sqlite_file: a path to an sqlite3 file (only used if uri_conn is None)
    :param str uri_conn: a sqlalchemy connection string
    :param str currency: the ISO symbol of the default currency of the book
    :param bool overwrite: True if book should be deleted and recreated if it exists already
    :param bool keep_foreign_keys: True if the foreign keys should be kept (may not work at all with GnuCash)
    :param str db_type: type of database in ["postgres","mysql"]
    :param str db_user: username of database
    :param str db_password: password for the use of database
    :param str db_name: name of database
    :param str db_host: host of database
    :param int db_port: port of database
    :param bool check_same_thread: sqlite flag that restricts connection use to the thread that created (see False for use in ipython/flask/... but read first https://docs.python.org/3/library/sqlite3.html)
    :param str pg_template: the postgres template to use when creating the database. One of template1 or template0 (default template0). Irrelevant for other databases than postgres.
    :return: the document as a gnucash session
    :rtype: :class:`GncSession`

    :raises GnucashException: if document already exists and overwrite is False
    """
    from sqlalchemy_utils.functions import database_exists, create_database, drop_database

    VERSION_FORMAT = "3.0"

    uri_conn = build_uri(sqlite_file, uri_conn,
                         db_type, db_user, db_password, db_name, db_host, db_port,
                         check_same_thread,
                         )

    # create database (if DB is not a sqlite in memory)
    if uri_conn != "sqlite:///:memory:":
        if database_exists(uri_conn):
            if overwrite:
                drop_database(uri_conn)
            else:
                raise GnucashException("'{}' db already exists".format(uri_conn))
        create_database(uri_conn, template=pg_template)

    engine = create_piecash_engine(uri_conn, **kwargs)

    # drop constraints if we de not want to keep them (keep_foreign_keys=False), the default
    if not keep_foreign_keys:
        for n, tbl in DeclarativeBase.metadata.tables.items():
            # drop index constraints
            for idx in tbl.indexes:
                if idx.name.startswith("ix_") or idx.name.startswith("_"):
                    event.listen(tbl,
                                 "after_create",
                                 DropIndex(idx),
                                 once=True)
            # drop FK constraints
            for cstr in tbl.constraints:
                if isinstance(cstr, PrimaryKeyConstraint):
                    continue
                else:
                    event.listen(tbl,
                                 "before_drop",
                                 DropConstraint(cstr),
                                 once=True)
    #
    # create all (tables, fk, ...)
    DeclarativeBase.metadata.create_all(engine)

    s = Session(bind=engine)

    # create all rows in version table
    assert VERSION_FORMAT in version_supported, "The 'version_format'={} is not supported. " \
                                                "Choose one of {}".format(VERSION_FORMAT,
                                                                          list(version_supported.keys()))
    for table_name, table_version in version_supported[VERSION_FORMAT].items():
        s.add(Version(table_name=table_name, table_version=table_version))

    # create book and merge with session

    b = Book()
    s.add(b)
    adapt_session(s, book=b, readonly=False)

    # create commodities and initial accounts
    from .account import Account

    b.root_account = Account(name="Root Account", type="ROOT",
                             commodity=factories.create_currency_from_ISO(currency),
                             book=b)
    b.root_template = Account(name="Template Root", type="ROOT", commodity=None, book=b)
    b.save()

    return b


def open_book(sqlite_file=None,
              uri_conn=None,
              readonly=True,
              open_if_lock=False,
              do_backup=True,
              db_type=None,
              db_user=None,
              db_password=None,
              db_name=None,
              db_host=None,
              db_port=None,
              check_same_thread=True,
              check_exists=True,
              **kwargs):
    """Open an existing GnuCash book

    :param str sqlite_file: a path to an sqlite3 file (only used if uri_conn is None)
    :param str uri_conn: a sqlalchemy connection string
    :param bool readonly: open the file as readonly (useful to play with and avoid any unwanted save)
    :param bool open_if_lock: open the file even if it is locked by another user
        (using open_if_lock=True with readonly=False is not recommended)
    :param bool do_backup: do a backup if the file written in RW (i.e. readonly=False)
        (this only works with the sqlite backend and copy the file with .{:%Y%m%d%H%M%S}.gnucash appended to it)
    :param str db_type: type of database in ["postgres","mysql"]
    :param str db_user: username of database
    :param str db_password: password for the use of database
    :param str db_name: name of database
    :param str db_host: host of database
    :param str db_port: port of database
    :param bool check_same_thread: sqlite flag that restricts connection use to the thread that created (see False for use in ipython/flask/... but read first https://docs.python.org/3/library/sqlite3.html)
    :param bool check_exists: check if the database exists before connecting

    :return: the document as a gnucash session
    :rtype: :class:`GncSession`
    :raises GnucashException: if the document does not exist
    :raises GnucashException: if there is a lock on the file and open_if_lock is False

    """
    uri_conn = build_uri(sqlite_file, uri_conn,
                         db_type, db_user, db_password, db_name, db_host, db_port,
                         check_same_thread,
                         )

    if uri_conn == "sqlite:///:memory:":
        raise ValueError("An in memory sqlite gnucash databook cannot be opened, it should be created")

    # check if the database exists
    if check_exists and not database_exists(uri_conn):
            raise GnucashException("Database '{}' does not exist (please use create_book to create " \
                                   "GnuCash books from scratch). If you want to bypass this existence check, "
                                   "use the argument check_exists=False.".format(uri_conn))

    engine = create_piecash_engine(uri_conn, **kwargs)

    # backup database if readonly=False and do_backup=True
    if not readonly and do_backup:
        if engine.name != "sqlite":
            raise GnucashException(
                "Cannot do a backup for engine '{}'. Do yourself a backup and then specify do_backup=False".format(
                    engine.name))

        url = uri_conn[len("sqlite:///"):].replace("?check_same_thread=False", "")
        url_backup = url + ".{:%Y%m%d%H%M%S}.gnucash".format(datetime.datetime.now())

        shutil.copyfile(url, url_backup)

    locks = list(engine.execute(gnclock.select()))

    # ensure the file is not locked by GnuCash itself
    if locks and not open_if_lock:
        raise GnucashException("Lock on the file")

    s = Session(bind=engine)

    # check the versions in the table versions is consistent with the API
    version_book = {v.table_name: v.table_version
                    for v in s.query(Version).all()
                    if "Gnucash" not in v.table_name}
    for version, vt in version_supported.items():
        if version_book == {k: v
                            for k, v in vt.items() if
                            "Gnucash" not in k}:
            break
    else:
        raise ValueError("Unsupported table versions")
    assert version == "3.0" or version == "3.7" or version == "4.1", "This version of piecash only support books from gnucash (3.0|3.7|4.1) " \
                             "which is not the case for {}".format(uri_conn)

    book = s.query(Book).one()
    adapt_session(s, book=book, readonly=readonly)

    return book


def adapt_session(session, book, readonly):
    """
    Change the SA session object to add some features.

    :param session: the SA session object that will be modified in place
    :param book: the gnucash singleton book linked to the SA session
    :param readonly: True if the session should not allow commits.
    :return:
    """
    # link session and book together
    book.session = session
    session.book = book
    book.uri = session.bind.url

    # def new_flush(*args, **kwargs):
    # if session.dirty or session.new or session.deleted:
    # session.rollback()
    # raise GnucashException("You cannot change the DB, it is locked !")

    # add logic to make session readonly
    def readonly_commit(*args, **kwargs):
        # session.rollback()
        raise GnucashException("You cannot change the DB, it was opened as readonly!")

    if readonly:
        session.commit = readonly_commit

    # add logic to create/delete GnuCash locks
    def delete_lock():
        session.execute(gnclock.delete(whereclause=(gnclock.c.hostname == socket.gethostname())
                                                   and (gnclock.c.pid == os.getpid())))
        session.commit()

    session.delete_lock = delete_lock

    def create_lock():
        session.execute(gnclock.insert(values=dict(hostname=socket.gethostname(), pid=os.getpid())))
        session.commit()

    session.create_lock = create_lock

    # add logic to track if a session has been modified or not
    session._is_modified = False
    session._all_changes = {}

    @event.listens_for(session, 'after_flush')
    def receive_after_flush(session, flush_context):
        session._is_modified = not session.is_saved

    @event.listens_for(session, 'after_commit')
    @event.listens_for(session, 'after_rollback')
    def init_session_status(session, *args, **kwargs):
        session._is_modified = False
        session._all_changes.clear()

    session.book.session_changes = defaultdict(list)

    session.__class__.is_saved = property(
        fget=lambda self: not (self._is_modified or self.dirty or self.deleted or self.new),
        doc="True if nothing has yet been changed (False otherwise)")


event.listen(Session, 'before_commit', Book.validate_book)
event.listen(Session, 'before_flush', Book.track_dirty)



from __future__ import print_function
from piecash import create_book

# create by default an in memory sqlite version
with create_book(echo=False) as book:

    print("Book is saved:", book.is_saved, end=' ')
    print(" ==> book description:", book.root_account.description)

    print("changing description...")
    book.root_account.description = "hello, book"
    print("Book is saved:", book.is_saved, end=' ')
    print(" ==> book description:", book.root_account.description)

    print("saving...")
    book.save()

    print("Book is saved:", book.is_saved, end=' ')
    print(" ==> book description:", book.root_account.description)

    print("changing description...")
    book.root_account.description = "nevermind, book"
    print("Book is saved:", book.is_saved, end=' ')
    print(" ==> book description:", book.root_account.description)

    print("cancel...")
    book.cancel()

    print("Book is saved:", book.is_saved, end=' ')
    print(" ==> book description:", book.root_account.description)



# coding=utf-8
from __future__ import unicode_literals

from datetime import datetime
from decimal import Decimal

import pytest
import tzlocal

from piecash import GnucashException, Commodity
from piecash.core import factories
from test_helper import db_sqlite_uri, db_sqlite, new_book, new_book_USD, book_uri, book_basic, needweb

# dummy line to avoid removing unused symbols

a = db_sqlite_uri, db_sqlite, new_book, new_book_USD, book_uri, book_basic


class TestFactoriesCommodities(object):
    def test_create_stock_accounts_simple(self, book_basic):
        with pytest.raises(GnucashException):
            factories.create_stock_accounts(
                book_basic.default_currency, broker_account=book_basic.accounts(name="broker")
            )

        broker = book_basic.accounts(name="broker")
        appl = Commodity(namespace="NMS", mnemonic="AAPL", fullname="Apple")
        acc, inc_accounts = factories.create_stock_accounts(appl, broker_account=broker)

        assert inc_accounts == []
        assert broker.children == [acc]

    def test_create_stock_accounts_incomeaccounts(self, book_basic):
        broker = book_basic.accounts(name="broker")
        income = book_basic.accounts(name="inc")

        appl = Commodity(namespace="NMS", mnemonic="AAPL", fullname="Apple")
        appl["quoted_currency"] = "USD"
        acc, inc_accounts = factories.create_stock_accounts(
            appl, broker_account=broker, income_account=income, income_account_types="D"
        )
        assert len(inc_accounts) == 1

        acc, inc_accounts = factories.create_stock_accounts(
            appl, broker_account=broker, income_account=income, income_account_types="CL"
        )
        assert len(inc_accounts) == 1
        acc, inc_accounts = factories.create_stock_accounts(
            appl, broker_account=broker, income_account=income, income_account_types="CS"
        )
        assert len(inc_accounts) == 1
        acc, inc_accounts = factories.create_stock_accounts(
            appl, broker_account=broker, income_account=income, income_account_types="I"
        )
        assert len(inc_accounts) == 1
        acc, inc_accounts = factories.create_stock_accounts(
            appl, broker_account=broker, income_account=income, income_account_types="D/CL/CS/I"
        )
        assert len(income.children) == 4
        book_basic.flush()
        assert sorted(income.children, key=lambda x: x.guid) == sorted(
            [_acc.parent for _acc in inc_accounts], key=lambda x: x.guid
        )
        assert broker.children == [acc]

    @needweb
    def test_create_stock_from_symbol(self, book_basic):
        assert len(book_basic.commodities) == 2

        factories.create_stock_from_symbol("AAPL", book_basic)

        assert len(book_basic.commodities) == 3

        cdty = book_basic.commodities(mnemonic="AAPL")

        assert cdty.namespace == "NMS"
        assert cdty.quote_tz == "America/New_York"
        assert cdty.quote_source == "yahoo"
        assert cdty.mnemonic == "AAPL"
        assert cdty.fullname == "Apple Inc."

    def test_create_currency_from_ISO(self, book_basic):
        assert factories.create_currency_from_ISO("CAD").fullname == "Canadian Dollar"

        with pytest.raises(ValueError):
            factories.create_currency_from_ISO("EFR").fullname


class TestFactoriesTransactions(object):
    def test_single_transaction(self, book_basic):
        today = datetime.today()
        print("today=", today)
        factories.single_transaction(
            today.date(),
            today,
            "my test",
            Decimal(100),
            from_account=book_basic.accounts(name="inc"),
            to_account=book_basic.accounts(name="asset"),
        )
        book_basic.save()
        tr = book_basic.transactions(description="my test")
        assert len(tr.splits) == 2
        sp1, sp2 = tr.splits
        if sp1.value > 0:
            sp2, sp1 = sp1, sp2
        # sp1 has negative value
        assert sp1.account == book_basic.accounts(name="inc")
        assert sp2.account == book_basic.accounts(name="asset")
        assert sp1.value == -sp2.value
        assert sp1.quantity == sp1.value
        assert tr.enter_date == tzlocal.get_localzone().localize(today.replace(microsecond=0))
        assert tr.post_date == tzlocal.get_localzone().localize(today).date()

    def test_single_transaction_tz(self, book_basic):
        today = tzlocal.get_localzone().localize(datetime.today())
        tr = factories.single_transaction(
            today.date(),
            today,
            "my test",
            Decimal(100),
            from_account=book_basic.accounts(name="inc"),
            to_account=book_basic.accounts(name="asset"),
        )
        book_basic.save()
        tr = book_basic.transactions(description="my test")
        assert tr.post_date == today.date()
        assert tr.enter_date == today.replace(microsecond=0)

    def test_single_transaction_rollback(self, book_basic):
        today = tzlocal.get_localzone().localize(datetime.today())
        factories.single_transaction(
            today.date(),
            today,
            "my test",
            Decimal(100),
            from_account=book_basic.accounts(name="inc"),
            to_account=book_basic.accounts(name="asset"),
        )
        book_basic.validate()
        assert len(book_basic.transactions) == 1
        book_basic.cancel()
        assert len(book_basic.transactions) == 0



from decimal import Decimal

from sqlalchemy import Column, VARCHAR, INTEGER, BIGINT, ForeignKey, and_, event
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.orm import composite, relation, foreign

from .._common import hybrid_property_gncnumeric, CallableList
from .._declbase import DeclarativeBaseGuid
from ..sa_extra import ChoiceType

TaxIncludedType = [
    (1, "YES"),
    (2, "NO"),
    (3, "USEGLOBAL")
]


class Address(object):
    """An Address object encapsulates information regarding an address in GnuCash.

    Attributes:
        name (str): self explanatory
        addr1 (str): self explanatory
        addr2 (str): self explanatory
        addr3 (str): self explanatory
        addr4 (str): self explanatory
        email (str): self explanatory
        fax (str): self explanatory
        phone (str): self explanatory
    """
    _address_fields = ['name', 'addr1', 'addr2', 'addr3', 'addr4', 'email', 'fax', 'phone']

    def __init__(self, name="", addr1="", addr2="", addr3="", addr4="", email="", fax="", phone=""):
        self.name = name
        self.addr1 = addr1
        self.addr2 = addr2
        self.addr3 = addr3
        self.addr4 = addr4
        self.email = email
        self.fax = fax
        self.phone = phone

    def __composite_values__(self):
        return (getattr(self, fld) for fld in Address._address_fields)

    def __eq__(self, other):
        return isinstance(other, Address) and all(
            getattr(other, fld) == getattr(self, fld) for fld in Address._address_fields)

    def __ne__(self, other):
        return not self.__eq__(other)


class Person:
    """A mixin declaring common field for Customer, Vendor and Employee"""
    active = Column('active', INTEGER(), nullable=False)

    id = Column('id', VARCHAR(length=2048), nullable=False)

    addr_name = Column('addr_name', VARCHAR(length=1024))
    addr_addr1 = Column('addr_addr1', VARCHAR(length=1024))
    addr_addr2 = Column('addr_addr2', VARCHAR(length=1024))
    addr_addr3 = Column('addr_addr3', VARCHAR(length=1024))
    addr_addr4 = Column('addr_addr4', VARCHAR(length=1024))
    addr_phone = Column('addr_phone', VARCHAR(length=128))
    addr_fax = Column('addr_fax', VARCHAR(length=128))
    addr_email = Column('addr_email', VARCHAR(length=256))

    @declared_attr
    def address(cls):
        return composite(Address, cls.addr_name, cls.addr_addr1, cls.addr_addr2, cls.addr_addr3, cls.addr_addr4,
                         cls.addr_email, cls.addr_fax, cls.addr_phone)

    @declared_attr
    def currency_guid(cls):
        return Column('currency', VARCHAR(length=32), ForeignKey('commodities.guid'), nullable=False)

    @declared_attr
    def currency(cls):
        return relation('Commodity')

    # hold the name of the counter to use for id
    _counter_name = None

    def _assign_id(self):
        if not self.id:
            cnt = getattr(self.book, self._counter_name) + 1
            setattr(self.book, self._counter_name, cnt)
            self.id = "{:06d}".format(cnt)

    def object_to_validate(self, change):
        yield self

    def validate(self):
        if self.id is None:
            raise ValueError("You cannot flush a {} that is not added to a book".format(self.__class__.__name__))

    def on_book_add(self):
        self._assign_id()
        for job in self.jobs:
            job.on_book_add()

    def __str__(self):
        return "{}<{}:{}>".format(self.__class__.__name__, self.id, self.name)

    @classmethod
    def __declare_last__(cls):
        from .invoice import Job
        owner_type = PersonType.get(cls, None)
        if owner_type:
            cls.jobs = relation('Job',
                                primaryjoin=and_(
                                    cls.guid == foreign(Job.owner_guid),
                                    owner_type == Job.owner_type,
                                ),
                                cascade='all, delete-orphan',
                                collection_class=CallableList,
                                )

            @event.listens_for(cls.jobs, "append")
            def add(target, value, initiator):
                value.owner_type = owner_type
                value.owner_guid = target.guid
                value._assign_id()


class Customer(Person, DeclarativeBaseGuid):
    """
    A GnuCash Customer

    Attributes:
        name (str): name of the Customer
        id (str): autonumber id with 5 digits (initialised to book.counter_customer + 1)
        notes (str): notes
        active (int): 1 if the customer is active, 0 otherwise
        discount (:class:`decimal.Decimal`): see Gnucash documentation
        credit (:class:`decimal.Decimal`): see Gnucash documentation
        currency (:class:`piecash.core.commodity.Commodity`): the currency of the customer
        tax_override (int): 1 if tax override, 0 otherwise
        address (:class:`Address`): the address of the customer
        shipping_address (:class:`Address`): the shipping address of the customer
        tax_included (str): 'yes', 'no', 'use global'
        taxtable (:class:`piecash.business.tax.TaxTable`): tax table of the customer
        term (:class:`piecash.business.invoice.Billterm`): bill term of the customer
    """
    __tablename__ = 'customers'

    __table_args__ = {}

    # column definitions
    name = Column('name', VARCHAR(length=2048), nullable=False)
    # id is nullable as it is set during validation (happening after flush)
    notes = Column('notes', VARCHAR(length=2048), nullable=False)
    _discount_num = Column('discount_num', BIGINT(), nullable=False)
    _discount_denom = Column('discount_denom', BIGINT(), nullable=False)
    discount = hybrid_property_gncnumeric(_discount_num, _discount_denom)
    _credit_num = Column('credit_num', BIGINT(), nullable=False)
    _credit_denom = Column('credit_denom', BIGINT(), nullable=False)
    credit = hybrid_property_gncnumeric(_credit_num, _credit_denom)
    tax_override = Column('tax_override', INTEGER(), nullable=False)

    shipaddr_name = Column('shipaddr_name', VARCHAR(length=1024))
    shipaddr_addr1 = Column('shipaddr_addr1', VARCHAR(length=1024))
    shipaddr_addr2 = Column('shipaddr_addr2', VARCHAR(length=1024))
    shipaddr_addr3 = Column('shipaddr_addr3', VARCHAR(length=1024))
    shipaddr_addr4 = Column('shipaddr_addr4', VARCHAR(length=1024))
    shipaddr_phone = Column('shipaddr_phone', VARCHAR(length=128))
    shipaddr_fax = Column('shipaddr_fax', VARCHAR(length=128))
    shipaddr_email = Column('shipaddr_email', VARCHAR(length=256))
    shipping_address = composite(Address, shipaddr_name, shipaddr_addr1, shipaddr_addr2, shipaddr_addr3, shipaddr_addr4,
                                 shipaddr_email, shipaddr_fax, shipaddr_phone)

    term_guid = Column('terms', VARCHAR(length=32), ForeignKey('billterms.guid'))
    tax_included = Column('tax_included', ChoiceType(TaxIncludedType))
    taxtable_guid = Column('taxtable', VARCHAR(length=32), ForeignKey('taxtables.guid'))

    # relation definitions
    taxtable = relation('Taxtable')
    term = relation('Billterm')

    def __init__(self,
                 name,
                 currency,
                 id=None,
                 notes="",
                 active=1,
                 tax_override=0,
                 credit=Decimal(0),
                 discount=Decimal(0),
                 taxtable=None,
                 address=None,
                 shipping_address=None,
                 tax_included="USEGLOBAL",
                 book=None):
        self.name = name
        self.currency = currency
        self.notes = notes
        self.active = active
        self.credit = credit
        self.discount = discount
        self.tax_included = tax_included
        self.taxtable = taxtable
        self.tax_override = tax_override
        if address is None:
            address = Address(name=name)
        self.address = address
        if shipping_address is None:
            shipping_address = Address("")
        self.shipping_address = shipping_address

        if book and id is None:
            book.add(self)
        elif id is not None:
            if isinstance(id, int):
                self.id = str(id)
            else:
                self.id = id

    _counter_name = "counter_customer"


class Employee(Person, DeclarativeBaseGuid):
    """
    A GnuCash Employee

    Attributes:
        name (str): name of the Employee
        id (str): autonumber id with 5 digits (initialised to book.counter_employee + 1)
        language (str): language
        active (int): 1 if the employee is active, 0 otherwise
        workday (:class:`decimal.Decimal`): see Gnucash documentation
        rate (:class:`decimal.Decimal`): see Gnucash documentation
        currency (:class:`piecash.core.commodity.Commodity`): the currency of the employee
        address (:class:`Address`): the address of the employee
        creditcard_account (:class:`piecash.core.account.Account`): credit card account for the employee

    """
    __tablename__ = 'employees'

    __table_args__ = {}

    # column definitions
    name = Column('username', VARCHAR(length=2048), nullable=False)
    # id is nullable as it is set during validation (happening after flush)
    language = Column('language', VARCHAR(length=2048), nullable=False)
    acl = Column('acl', VARCHAR(length=2048), nullable=False)
    ccard_guid = Column('ccard_guid', VARCHAR(length=32), ForeignKey('accounts.guid'))
    _workday_num = Column('workday_num', BIGINT(), nullable=False)
    _workday_denom = Column('workday_denom', BIGINT(), nullable=False)
    workday = hybrid_property_gncnumeric(_workday_num, _workday_denom)
    _rate_num = Column('rate_num', BIGINT(), nullable=False)
    _rate_denom = Column('rate_denom', BIGINT(), nullable=False)
    rate = hybrid_property_gncnumeric(_rate_num, _rate_denom)

    # relation definitions
    creditcard_account = relation('Account')

    def __init__(self,
                 name,
                 currency,
                 creditcard_account=None,
                 id=None,
                 active=1,
                 acl="",
                 language="",
                 workday=Decimal(0),
                 rate=Decimal(0),
                 address=None,
                 book=None):
        self.name = name
        self.currency = currency
        self.active = active
        self.workday = workday
        self.rate = rate
        self.acl = acl
        self.language = language
        self.creditcard_account = creditcard_account
        if address is None:
            address = Address(name=name)
        self.address = address

        if book and id is None:
            book.add(self)

        elif id is not None:
            if isinstance(id, int):
                self.id = str(id)
            else:
                self.id = id

    _counter_name = "counter_employee"

    def on_book_add(self):
        self._assign_id()


class Vendor(Person, DeclarativeBaseGuid):
    """
    A GnuCash Vendor

    Attributes:
        name (str): name of the Vendor
        id (str): autonumber id with 5 digits (initialised to book.counter_vendor + 1)
        notes (str): notes
        active (int): 1 if the vendor is active, 0 otherwise
        currency (:class:`piecash.core.commodity.Commodity`): the currency of the vendor
        tax_override (int): 1 if tax override, 0 otherwise
        address (:class:`Address`): the address of the vendor
        tax_included (str): 'YES', 'NO', 'USEGLOBAL'
        taxtable (:class:`piecash.business.tax.TaxTable`): tax table of the vendor
        term (:class:`piecash.business.invoice.Billterm`): bill term of the vendor
    """
    __tablename__ = 'vendors'

    __table_args__ = {}

    # column definitions
    name = Column('name', VARCHAR(length=2048), nullable=False)
    # id is nullable as it is set during validation (happening after flush)
    notes = Column('notes', VARCHAR(length=2048), nullable=False)
    tax_override = Column('tax_override', INTEGER(), nullable=False)

    term_guid = Column('terms', VARCHAR(length=32), ForeignKey('billterms.guid'))
    tax_included = Column('tax_inc', VARCHAR(length=2048))
    tax_table_guid = Column('tax_table', VARCHAR(length=32), ForeignKey('taxtables.guid'))

    # relation definitions
    taxtable = relation('Taxtable')
    term = relation('Billterm')

    def __init__(self,
                 name,
                 currency,
                 id=None,
                 notes="",
                 active=1,
                 tax_override=0,
                 taxtable=None,
                 credit=Decimal(0),
                 discount=Decimal(0),
                 address=None,
                 tax_included="USEGLOBAL",
                 book=None):
        self.name = name
        self.currency = currency
        self.notes = notes
        self.active = active
        self.credit = credit
        self.discount = discount
        self.tax_included = tax_included
        self.taxtable = taxtable
        self.tax_override = tax_override
        if address is None:
            address = Address(name=name)
        self.address = address

        if book and id is None:
            book.add(self)
        elif id is not None:
            if isinstance(id, int):
                self.id = str(id)
            else:
                self.id = id

    _counter_name = "counter_vendor"


PersonType = {
    Vendor: 4,
    Customer: 2
}



# -*- coding: latin-1 -*-
import os
from datetime import date
from pathlib import PurePath, Path

import pytest
from sqlalchemy_utils import database_exists, drop_database

from piecash import (
    create_book,
    open_book,
    Account,
    Commodity,
    Employee,
    Customer,
    Vendor,
    Transaction,
    Split,
    Price,
)

test_folder = Path(__file__).parent
# book_folder = test_folder / ".." / "gnucash_books"
book_folder = test_folder / "books"
file_template = book_folder / "empty_book.gnucash"
file_template = book_folder / "default_3_0_0_basic.gnucash"
file_for_test = test_folder / "empty_book_for_test.gnucash"
file_template_full = book_folder / "test_book.gnucash"
# file_template_full = book_folder / "reference" / "3_0" / "default_3_0_0_full_options.gnucash"
file_template_full = book_folder / "all-accounts.gnucash"
file_for_test_full = test_folder / "test_book_for_test.gnucash"
file_ghost_kvp_scheduled_transaction = book_folder / "ghost_kvp_scheduled_transaction.gnucash"
file_ghost_kvp_scheduled_transaction_for_test = (
        test_folder / "ghost_kvp_scheduled_transaction_for_test.gnucash"
)


def run_file(fname):
    with open(fname) as f:
        code = compile(f.read(), fname, "exec")
        exec(code, {})


db_sqlite = test_folder / "foozbar.sqlite"

TRAVIS = os.environ.get("TRAVIS", False)
APPVEYOR = os.environ.get("APPVEYOR", False)
LOCALSERVER = os.environ.get("PIECASH_DBSERVER_TEST", False)
LOCALSERVER_USERNAME = os.environ.get("PIECASH_DBSERVER_TEST_USERNAME", "")

db_sqlite_uri = "sqlite:///{}".format(db_sqlite)
databases_to_check = [None, db_sqlite_uri]
db_config = {"sqlite": dict(sqlite_file=db_sqlite), "sqlite_in_mem": dict(sqlite_file=None)}

if TRAVIS:
    pg_password = os.environ.get("PG_PASSWORD", "")
    db_user = "travis"
    databases_to_check.append(
        "postgresql://postgres:{pwd}@localhost:5432/foo".format(pwd=pg_password)
    )
    databases_to_check.append("mysql+pymysql://travis:@localhost/foo?charset=utf8")
    db_config.update(
        {
            "postgres": dict(
                db_type="postgres",
                db_name="foo",
                db_user="postgres",
                db_password=pg_password,
                db_host="localhost",
                db_port=5432,
            ),
            "mysql": dict(
                db_type="mysql",
                db_name="foo",
                db_user="travis",
                db_password="",
                db_host="localhost",
                db_port=3306,
            ),
        }
    )
elif LOCALSERVER:
    pg_password = os.environ.get("PG_PASSWORD", "")
    pg_port = os.environ.get("PIECASH_DBSERVER_TEST_PORT", "5432")
    db_user = "travis"
    databases_to_check.append(
        "postgresql://{username}:{pwd}@localhost:{pg_port}/foo".format(
            username=LOCALSERVER_USERNAME, pwd=pg_password, pg_port=pg_port
        )
    )
    db_config.update(
        {
            "postgres": dict(
                db_type="postgres",
                db_name="foo",
                db_user=LOCALSERVER_USERNAME,
                db_password=pg_password,
                db_host="localhost",
                db_port=pg_port,
            )
        }
    )
elif APPVEYOR:
    databases_to_check.append("postgresql://postgres:Password12!@localhost:5432/foo")
    databases_to_check.append("mysql+pymysql://root:Password12!@localhost/foo?charset=utf8")
    db_config.update(
        {
            "postgres": dict(
                db_type="postgres",
                db_name="foo",
                db_user="postgres",
                db_password="Password12!",
                db_host="localhost",
                db_port=5432,
            ),
            "mysql": dict(
                db_type="mysql",
                db_name="foo",
                db_user="root",
                db_password="Password12!",
                db_host="localhost",
                db_port=3306,
            ),
        }
    )
else:
    pass


@pytest.yield_fixture(params=[Customer, Vendor, Employee])
def Person(request):
    yield request.param


@pytest.yield_fixture(params=db_config.items())
def book_db_config(request):
    from piecash.core.session import build_uri

    sql_backend, db_config = request.param
    name = build_uri(**db_config)

    if sql_backend != "sqlite_in_mem" and database_exists(name):
        drop_database(name)

    yield db_config

    if sql_backend != "sqlite_in_mem" and database_exists(name):
        drop_database(name)


@pytest.yield_fixture(params=databases_to_check[1:])
def book_uri(request):
    name = request.param

    if name and database_exists(name):
        drop_database(name)
    yield name

    if name and database_exists(name):
        drop_database(name)


@pytest.yield_fixture(params=databases_to_check)
def new_book(request):
    name = request.param

    if name and database_exists(name):
        drop_database(name)

    with create_book(uri_conn=name, keep_foreign_keys=False) as b:
        yield b

    if name and database_exists(name):
        drop_database(name)


@pytest.yield_fixture(params=databases_to_check)
def new_book_USD(request):
    name = request.param

    if name and database_exists(name):
        drop_database(name)

    with create_book(uri_conn=name, currency="USD", keep_foreign_keys=False) as b:
        yield b

    if name and database_exists(name):
        drop_database(name)


@pytest.yield_fixture(params=databases_to_check)
def book_basic(request):
    name = request.param

    if name and database_exists(name):
        drop_database(name)
    # create new book
    with create_book(uri_conn=name, currency="EUR", keep_foreign_keys=False) as b:
        # create some accounts
        curr = b.currencies[0]
        cdty = Commodity(namespace=u"échange", mnemonic=u"ďoŕ", fullname=u"Example of unicode déta")
        a = Account(name="asset", type="ASSET", commodity=curr, parent=b.root_account)
        Account(name="broker", type="STOCK", commodity=cdty, parent=a)
        Account(name="exp", type="EXPENSE", commodity=curr, parent=b.root_account)
        Account(name="inc", type="INCOME", commodity=curr, parent=b.root_account)
        b.flush()

        yield b

    if name and database_exists(name):
        drop_database(name)


@pytest.yield_fixture(params=databases_to_check)
def book_transactions(request):
    name = request.param

    if name and database_exists(name):
        drop_database(name)
    # create new book
    with create_book(uri_conn=name, currency="EUR", keep_foreign_keys=False) as b:
        # create some accounts
        curr = b.default_currency
        other_curr = b.currencies(mnemonic="USD")
        cdty = Commodity(
            namespace=u"BEL20", mnemonic=u"GnuCash Inc.", fullname=u"GnuCash Inc. stock"
        )
        asset = Account(name="asset", type="ASSET", commodity=curr, parent=b.root_account)
        foreign_asset = Account(
            name="foreign asset", type="ASSET", commodity=other_curr, parent=b.root_account
        )
        stock = Account(name="broker", type="STOCK", commodity=cdty, parent=asset)
        expense = Account(name="exp", type="EXPENSE", commodity=curr, parent=b.root_account)
        income = Account(name="inc", type="INCOME", commodity=curr, parent=b.root_account)

        tr1 = Transaction(
            post_date=date(2015, 10, 21),
            description="my revenue",
            currency=curr,
            splits=[Split(account=asset, value=(1000, 1)), Split(account=income, value=(-1000, 1))],
        )
        tr2 = Transaction(
            post_date=date(2015, 10, 25),
            description="my expense",
            currency=curr,
            splits=[
                Split(account=asset, value=(-100, 1)),
                Split(account=expense, value=(20, 1), memo="cost of X"),
                Split(account=expense, value=(80, 1), memo="cost of Y"),
            ],
        )
        tr_stock = Transaction(
            post_date=date(2015, 10, 29),
            description="my purchase of stock",
            currency=curr,
            splits=[
                Split(account=asset, value=(-200, 1)),
                Split(account=expense, value=(15, 1), memo="transaction costs"),
                Split(account=stock, value=(185, 1), quantity=(6, 1), memo="purchase of stock"),
            ],
        )
        tr_to_foreign = Transaction(
            post_date=date(2015, 10, 30),
            description="transfer to foreign asset",
            currency=curr,
            splits=[
                Split(account=asset, value=(-200, 1)),
                Split(account=foreign_asset, value=(200, 1), quantity=(135, 1)),
            ],
        )
        tr_from_foreign = Transaction(
            post_date=date(2015, 10, 31),
            description="transfer from foreign asset",
            currency=other_curr,
            splits=[
                Split(account=asset, value=(135, 1), quantity=(215, 1)),
                Split(account=foreign_asset, value=(-135, 1)),
            ],
        )
        Price(commodity=cdty, currency=other_curr, date=date(2015, 11, 1), value=(123, 100))
        Price(commodity=cdty, currency=other_curr, date=date(2015, 11, 4), value=(127, 100))
        Price(commodity=cdty, currency=curr, date=date(2015, 11, 2), value=(234, 100))

        b.save()
        yield b

    if name and database_exists(name):
        drop_database(name)


@pytest.yield_fixture()
def book_invoices(request):
    """
    Returns the book that contains invoices.
    """
    # name = request.param
    # print(name)
    file_template_full = book_folder / "invoices.gnucash"

    with open_book(file_template_full) as book:
        yield book


@pytest.yield_fixture(params=["", ".272"])
def book_sample(request):
    """
    Returns a simple sample book for 2.6.N
    """
    file_template_full = book_folder / "simple_sample{}.gnucash".format(request.param)

    with open_book(file_template_full) as book:
        yield book


def is_inmemory_sqlite(book_basic):
    # print book_basic.uri, book_basic.uri.get_dialect(), book_basic.uri.database, type(book_basic.uri), dir(book_basic.uri)
    # print "sqlite" in book_basic.uri and ":memory:" in book_basic.uri
    # fdsfdssfd
    return book_basic.uri.database == ":memory:"


needweb = pytest.mark.skipif(
    not (os.environ.get("DOGOONWEB", "False") == "True"), reason="no access to web"
)


def generate_book_fixture(filename):
    @pytest.yield_fixture(scope="module")
    def my_fixture():
        file_template = book_folder / filename

        with open_book(file_template) as book:
            yield book

    return my_fixture


book_reference_3_0_0_fulloptions = generate_book_fixture(
    PurePath() / "default_3_0_0_full_options.gnucash"
)

book_reference_3_0_0_basic = generate_book_fixture(PurePath() / "default_3_0_0_basic.gnucash")

# complex 2.6 book sample
book_complex = generate_book_fixture(PurePath() / "complex_sample.gnucash")

book_investment = generate_book_fixture(PurePath() / "investment.gnucash")



import csv
from pathlib import Path

from piecash import open_book

fields = [
    "DATE",
    "TRANSACTION VALUE",
    "DEBIT/CREDIT INDICATOR",
    "ACCOUNT",
    "ACCOUNT CODE",
    "CONTRA ACCOUNT",
    "CONTRA ACCOUNT CODE",
    "ENTRY TEXT",
]

GNUCASH_BOOK = "../gnucash_books/simple_sample.gnucash"
CSV_EXPORT = "export.csv"
REPORTING_YEAR = 2019

# open the book and the export file
with open_book(GNUCASH_BOOK, readonly=True, open_if_lock=True) as mybook, Path(CSV_EXPORT).open(
    "w", newline=""
) as f:
    # initialise the CSV writer
    csv_writer = csv.DictWriter(f, fieldnames=fields)
    csv_writer.writeheader()

    # iterate on all the transactions in the book
    for transaction in mybook.transactions:
        # filter transactions not in REPORTING_YEAR
        if transaction.post_date.year != REPORTING_YEAR:
            continue

        # handle only transactions with 2 splits
        if len(transaction.splits) != 2:
            print(
                f"skipping transaction {transaction} as it has more"
                f" than 2 splits in the transaction, dunno what to export to CSV"
            )
            continue

        # assign the two splits of the transaction
        split_one, split_two = transaction.splits
        # build the dictionary with the data of the transaction
        data = dict(
            zip(
                fields,
                [
                    transaction.post_date,
                    split_one.value,
                    split_one.is_debit,
                    split_one.account.name,
                    split_one.account.code,
                    split_two.account.name,
                    split_two.account.code,
                    transaction.description,
                ],
            )
        )
        # write the transaction to the CSV
        csv_writer.writerow(data)



# coding=utf-8
from __future__ import unicode_literals

import locale
import sys
from contextlib import contextmanager

import pytest


from piecash import create_book, Account, open_book
from piecash._common import get_system_currency_mnemonic
from piecash.core.session import build_uri
from test_helper import db_sqlite_uri, db_sqlite, new_book, new_book_USD, book_uri, book_db_config

# dummy line to avoid removing unused symbols
a = db_sqlite_uri, db_sqlite, new_book, new_book_USD, book_uri, book_db_config


class TestSession_create_book(object):
    def test_create_default(self, book_db_config):
        with create_book(keep_foreign_keys=False, **book_db_config) as b:
            a = Account(commodity=b.currencies(mnemonic="SEK"),
                        parent=b.root_account,
                        name="léviö",
                        type="ASSET")
            assert str(b.uri) == build_uri(**book_db_config)
            b.save()

        # reopen the DB except if sqlite_file is None
        if book_db_config.get("sqlite_file", True):
            with open_book(**book_db_config) as b:
                assert b.accounts(name="léviö").commodity == b.currencies(mnemonic="SEK")

    def test_build_uri(self):
        assert build_uri() == "sqlite:///:memory:"
        assert build_uri(sqlite_file="foo") == "sqlite:///foo"
        assert build_uri(uri_conn="sqlite:///foo") == "sqlite:///foo"
        with pytest.raises(ValueError):
            build_uri(db_name="f")

        with pytest.raises(KeyError):
            build_uri(db_type="pg",
                      db_user="foo",
                      db_password="pp",
                      db_name="pqsd",
                      db_host="qsdqs",
                      db_port=3434)

        assert build_uri(db_type="postgres",
                         db_user="foo",
                         db_password="pp",
                         db_name="pqsd",
                         db_host="qsdqs",
                         db_port=3434) == "postgresql://foo:pp@qsdqs:3434/pqsd"

        assert build_uri(db_type="mysql",
                         db_user="foo",
                         db_password="pp",
                         db_name="pqsd",
                         db_host="qsdqs",
                         db_port=3434) == "mysql+pymysql://foo:pp@qsdqs:3434/pqsd?charset=utf8"

        ### Test duplicate protocol spec. This happens when the open_book is called
        ### from GnuCash reports (.scm), gnucash-utilities.
        sqlite_uri = "sqlite:///some_file"
        uri = "sqlite:///some_file"
        assert build_uri(sqlite_file=uri) == sqlite_uri
        # When run with just the name (without sqlite:// prefix):
        uri = "some_file"
        assert build_uri(sqlite_file=uri) == sqlite_uri


@contextmanager
def locale_ctx(l):
    _l = locale.getlocale()

    locale.setlocale(locale.LC_ALL, l)

    yield

    locale.setlocale(locale.LC_ALL, _l)


if sys.platform == "win32":
    locales = {
        'English_United States.1252': "USD",
        'French_France.1252': "EUR",
    }
else:
    locales = {
        'en_US.UTF-8': "USD",
        'fr_FR.UTF-8': "EUR",
    }


@pytest.yield_fixture(params=locales)
def locale_set(request):
    yield request.param

def test_get_system_currency_mnemonic(locale_set):
    result = locales[locale_set]
    with locale_ctx(locale_set):
        assert get_system_currency_mnemonic() == result



from sqlalchemy import Column, INTEGER

from piecash.sa_extra import DeclarativeBase


class B:
    __tablename__ = "b"
    guid = Column(type_=INTEGER,primary_key=True)

    def __init__(self):
        pass

    def __unirepr__(self):
        return "B<>"




import shutil

from piecash import Invoice

#shutil.copy("invoices.gnucash", "test_vendor_exp.gnucash")

import piecash


with piecash.open_book("test_vendor_exp.gnucash", readonly=False) as book:
    for i, v in enumerate(book.customers):
        v.name = f"customer-{i}"
    for i, v in enumerate(book.vendors):
        v.name = f"vendor-{i}"

    #book.invoices[0].owner=book.vendors[0]
    #book.save()

with piecash.open_book("test_vendor_exp.gnucash", readonly=True) as book:
    for i,v in enumerate(book.vendors):
        print(v.name)
    for v in book.invoices:
        print(v.owner_guid)
        print(v.owner)



#!/usr/bin/env python

import click

from piecash.scripts.cli import cli


@cli.command()
@click.argument('book', type=click.Path(exists=True))
@click.argument('entities', type=click.Choice(['customers', 'vendors', 'prices']))
@click.option('--output', type=click.File('w'), default="-",
              help="File to which to export the data (default=stdout)")
@click.option('--inactive', is_flag=True, default=False,
              help="Include inactive entities (for vendors and customers)")
def export(book, entities, output, inactive):
    """Exports GnuCash ENTITIES.

    This scripts export ENTITIES from the BOOK in a CSV format.
    When possible, it exports in a format that can be used to import the data into GnuCash.

    \b
    Remarks:
    - for customers and vendors, the format does not include an header
    - for prices, the format can be used with the `piecash import` command.
    """
    from piecash import open_book

    with open_book(book, open_if_lock=True) as book:
        if entities == "prices":
            output.write("date,type,value,value_num, value_denom, currency,commodity,source\n")
            output.writelines(
                "{p.date:%Y-%m-%d},{p.type},{p.value},{p._value_num},{p._value_denom},{p.currency.mnemonic},{p.commodity.mnemonic},{p.source}\n".format(
                    p=p) for p in book.prices)

        elif entities in ["customers", "vendors"]:
            columns = "id, name, addr_name, addr_addr1, addr_addr2, addr_addr3, addr_addr4, " \
                      "addr_phone, addr_fax, addr_email, notes, shipaddr_name, " \
                      "shipaddr_addr1, shipaddr_addr2, shipaddr_addr3, shipaddr_addr4, " \
                      "shipaddr_phone, shipaddr_fax, shipaddr_email".split(", ")
            separator = ";"

            filter_entity = (lambda e: True) if inactive else (lambda e: e.active)

            # open the book
            res = "\n".join([separator.join(getattr(v, fld, "")
                                            for fld in columns)
                             for v in getattr(book, entities)
                             if filter_entity(v)
                             ])

            output.write(res)



# -*- coding: latin-1 -*-
import pytest

from piecash import Account, Commodity
from test_helper import db_sqlite_uri, db_sqlite, new_book, new_book_USD, book_uri

# dummy line to avoid removing unused symbols

a = db_sqlite_uri, db_sqlite, new_book, new_book_USD, book_uri


class TestAccount_create_account(object):
    def test_create_parentless_account(self, new_book):
        EUR = new_book.commodities[0]
        racc = new_book.root_account

        # create an account without parent that is not ROOT
        acc = Account(name="test account", type="ASSET", commodity=EUR)
        new_book.add(acc)
        with pytest.raises(ValueError):
            new_book.validate()
        new_book.cancel()

        # create an account without parent that is ROOT but with wrong name
        acc = Account(name="test account", type="ROOT", commodity=EUR)
        new_book.add(acc)
        with pytest.raises(ValueError):
            new_book.validate()
        new_book.cancel()

        # create an account without parent that is ROOT with correct name
        acc = Account(name="Root Account", type="ROOT", commodity=EUR)
        new_book.add(acc)
        new_book.flush()

        assert len(new_book.accounts) == 0
        root_accs = new_book.query(Account).all()
        assert len(root_accs) == 3

    def test_create_samenameandparent_accounts(self, new_book):
        EUR = new_book.commodities[0]
        racc = new_book.root_account
        acc1 = Account(name="test account", type="ASSET", commodity=EUR, parent=racc)
        acc2 = Account(name="test account", type="ASSET", commodity=EUR, parent=racc)
        with pytest.raises(ValueError):
            new_book.validate()

    def test_create_samenameanddifferentparent_accounts(self, new_book):
        EUR = new_book.commodities[0]
        racc = new_book.root_account

        # create 2 accounts with same name but different parent
        acc1 = Account(name="test account", type="ASSET", commodity=EUR, parent=racc)
        acc2 = Account(name="test account", type="ASSET", commodity=EUR, parent=acc1)
        new_book.flush()
        assert acc1.fullname == "test account"
        assert acc2.fullname == "test account:test account"

    def test_create_standardasset_account(self, new_book):
        EUR = new_book.commodities[0]
        racc = new_book.root_account

        # create normal account
        acc = Account(name="test account", type="ASSET", commodity=EUR, parent=racc)
        new_book.flush()
        assert len(new_book.accounts) == 1
        assert acc.non_std_scu == 0
        assert acc.commodity_scu == EUR.fraction
        assert acc.get_balance() == 0
        assert acc.sign == 1
        assert not acc.is_template

    def test_create_standardliability_account(self, new_book):
        EUR = new_book.commodities[0]
        racc = new_book.root_account

        # create normal account
        acc = Account(name="test account", type="LIABILITY", commodity=EUR, parent=racc)
        new_book.flush()
        assert len(new_book.accounts) == 1
        assert acc.non_std_scu == 0
        assert acc.commodity_scu == EUR.fraction
        assert acc.get_balance() == 0
        assert acc.sign == -1
        assert not acc.is_template

    def test_create_unknowntype_account(self, new_book):
        EUR = new_book.commodities[0]
        racc = new_book.root_account

        # create account with unknown type
        with pytest.raises(ValueError):
            acc = Account(name="test account", type="FOO", commodity=EUR, parent=racc)
            new_book.validate()

    def test_create_nobook_account(self, new_book):
        USD = Commodity(namespace="FOO", mnemonic="BAZ", fullname="cuz")

        # create account with no book attachable to it
        with pytest.raises(ValueError):
            acc = Account(name="test account", type="ASSET", commodity=USD)
            new_book.flush()

    def test_create_children_accounts(self, new_book):
        EUR = new_book.commodities[0]
        racc = new_book.root_account

        # create account with unknown type
        acc = Account(name="test account", type="ASSET", commodity=EUR, parent=racc,
                      children=[Account(name="test sub-account", type="ASSET", commodity=EUR)])
        new_book.flush()
        assert len(acc.children) == 1

    def test_create_unicodename_account(self, new_book):
        EUR = new_book.commodities[0]
        racc = new_book.root_account

        # create normal account
        acc = Account(name=u"inouï étrange", type="ASSET", commodity=EUR, parent=racc)
        new_book.flush()
        assert len(new_book.accounts) == 1
        assert len(repr(acc)) >= 2

    def test_create_root_subaccount(self, new_book):
        EUR = new_book.commodities[0]
        racc = new_book.root_account

        # create root account should raise an exception
        acc = Account(name="subroot accout", type="ROOT", commodity=EUR, parent=racc)
        with pytest.raises(ValueError):
            new_book.validate()

        # except if we add the control_mode 'allow-root-subaccounts' to the book
        new_book.control_mode.append("allow-root-subaccounts")
        new_book.validate()

        assert len(new_book.accounts) == 1


class TestAccount_features(object):
    def test_sign_accounts(self, new_book):
        EUR = new_book.commodities[0]
        neg = "EQUITY,PAYABLE,LIABILITY,CREDIT,INCOME".split(",")
        pos = "STOCK,MUTUAL,EXPENSE,BANK,TRADING,CASH,ASSET,RECEIVABLE".split(",")
        all = neg + pos
        for acc in all:
            assert Account(name=acc, type=acc, commodity=EUR).sign == (1 if acc in pos else -1)

    def test_scu(self, new_book):
        EUR = new_book.commodities[0]
        acc = Account(name="test", type="ASSET", commodity=EUR)
        assert acc.commodity_scu == EUR.fraction
        assert not acc.non_std_scu
        acc.commodity_scu = 100
        assert acc.non_std_scu
        acc.commodity_scu = None
        assert acc.commodity_scu == EUR.fraction
        assert not acc.non_std_scu



from __future__ import unicode_literals

import uuid
from enum import Enum

from sqlalchemy import Column, VARCHAR, ForeignKey, INTEGER
from sqlalchemy.orm import relation, validates

from .._common import CallableList, GncConversionError
from .._declbase import DeclarativeBaseGuid
from ..sa_extra import mapped_to_slot_property

root_types = {"ROOT"}
asset_types = {'RECEIVABLE', 'MUTUAL', 'CASH', 'ASSET', 'BANK', 'STOCK'}
liability_types = {'CREDIT', 'LIABILITY', 'PAYABLE'}
income_types = {"INCOME"}
expense_types = {"EXPENSE"}
trading_types = {'TRADING'}
equity_types = {"EQUITY"}
# : the different types of accounts
ACCOUNT_TYPES = equity_types | income_types | expense_types | asset_types | liability_types | root_types | trading_types


class AccountType(Enum):
    root = "ROOT"
    receivable = "RECEIVABLE"
    mutual = "MUTUAL"
    cash = "CASH"
    asset = "ASSET"
    bank = "BANK"
    stock = "STOCK"
    credit = "CREDIT"
    liability = "LIABILITY"
    payable = "PAYABLE"
    income = "INCOME"
    expense = "EXPENSE"
    trading = "TRADING"
    equity = "EQUITY"


# types that are compatible with other types
incexp_types = income_types | expense_types
assetliab_types = asset_types | liability_types

# types according to the sign of their balance
positive_types = asset_types | expense_types | trading_types
negative_types = liability_types | income_types | equity_types


def _is_parent_child_types_consistent(type_parent, type_child, control_mode):
    """
    Return True if the child account is consistent with the parent account in terms of types, i.e.:

    1) if the parent is a root account, child can be anything but a root account
    2) if the child is a root account, it must have no parent account
    3) both parent and child are of the same family (asset, equity, income&expense, trading)

    Arguments
        type_parent(str): the type of the parent account
        type_child(str):  the type of the child account

    Returns
        True if both accounts are consistent, False otherwise
    """
    if type_parent in root_types:
        if "allow-root-subaccounts" in control_mode:
            return type_child in ACCOUNT_TYPES
        else:
            return type_child in (ACCOUNT_TYPES - root_types)

    if type_child in root_types:
        return (type_parent is None) or ("allow-root-subaccounts" in control_mode)

    for acc_types in (assetliab_types, equity_types, incexp_types, trading_types):
        if (type_child in acc_types) and (type_parent in acc_types):
            return True

    return False


class Account(DeclarativeBaseGuid):
    """
    A GnuCash Account which is specified by its name, type and commodity.

    Attributes:
        type (str): type of the Account
        sign (int): 1 for accounts with positive balances, -1 for accounts with negative balances
        code (str): code of the Account
        commodity (:class:`piecash.core.commodity.Commodity`): the commodity of the account
        commodity_scu (int): smallest currency unit for the account
        non_std_scu (int): 1 if the scu of the account is NOT the same as the commodity
        description (str): description of the account
        name (str): name of the account
        fullname (str): full name of the account (including name of parent accounts separated by ':')
        placeholder (int): 1 if the account is a placeholder (should not be involved in transactions)
        hidden (int): 1 if the account is hidden
        is_template (bool): True if the account is a template account (ie commodity=template/template)
        parent (:class:`Account`): the parent account of the account (None for the root account of a book)
        children (list of :class:`Account`): the list of the children accounts
        splits (list of :class:`piecash.core.transaction.Split`): the list of the splits linked to the account
        lots (list of :class:`piecash.business.Lot`): the list of lots to which the account is linked
        book (:class:`piecash.core.book.Book`): the book if the account is the root account (else None)
        budget_amounts (list of :class:`piecash.budget.BudgetAmount`): list of budget amounts of the account
        scheduled_transaction (:class:`piecash.core.transaction.ScheduledTransaction`): scheduled transaction linked to the account
    """
    __tablename__ = 'accounts'

    __table_args__ = {}

    # column definitions
    guid = Column('guid', VARCHAR(length=32), primary_key=True, nullable=False, default=lambda: uuid.uuid4().hex)
    name = Column('name', VARCHAR(length=2048), nullable=False)
    type = Column('account_type', VARCHAR(length=2048), nullable=False)
    commodity_guid = Column('commodity_guid', VARCHAR(length=32), ForeignKey('commodities.guid'))
    _commodity_scu = Column('commodity_scu', INTEGER(), nullable=False)
    _non_std_scu = Column('non_std_scu', INTEGER(), nullable=False)

    @property
    def non_std_scu(self):
        return self._non_std_scu

    @property
    def commodity_scu(self):
        return self._commodity_scu

    @commodity_scu.setter
    def commodity_scu(self, value):
        if value is None:
            self._non_std_scu = 0
            if self.commodity:
                value = self.commodity.fraction
            else:
                value = 0
        else:
            self._non_std_scu = 1

        self._commodity_scu = value

    parent_guid = Column('parent_guid', VARCHAR(length=32), ForeignKey('accounts.guid'))
    code = Column('code', VARCHAR(length=2048))
    description = Column('description', VARCHAR(length=2048))
    hidden = Column('hidden', INTEGER())
    _placeholder = Column('placeholder', INTEGER())
    placeholder = mapped_to_slot_property(_placeholder,
                                          slot_name="placeholder",
                                          slot_transform=lambda v: "true" if v else None)

    # relation definitions
    commodity = relation('Commodity', back_populates='accounts')
    children = relation('Account',
                        back_populates='parent',
                        cascade='all, delete-orphan',
                        collection_class=CallableList,
                        )
    parent = relation('Account',
                      back_populates='children',
                      remote_side=guid,
                      )
    splits = relation('Split',
                      back_populates='account',
                      cascade='all, delete-orphan',
                      collection_class=CallableList,
                      )
    lots = relation('Lot',
                    back_populates='account',
                    cascade='all, delete-orphan',
                    collection_class=CallableList,
                    )
    budget_amounts = relation('BudgetAmount',
                              back_populates='account',
                              cascade='all, delete-orphan',
                              collection_class=CallableList,
                              )
    scheduled_transaction = relation('ScheduledTransaction',
                                     back_populates='template_account',
                                     cascade='all, delete-orphan',
                                     uselist=False,
                                     )

    def __init__(self,
                 name,
                 type,
                 commodity,
                 parent=None,
                 description='',
                 commodity_scu=None,
                 hidden=0,
                 placeholder=0,
                 code='',
                 book=None,
                 children=None):
        book = book or (commodity and commodity.book) or (parent and parent.book)
        if not book:
            raise ValueError("Could not find a book to attach the account to")
        book.add(self)

        self.name = name
        self.commodity = commodity
        self.type = type
        self.parent = parent
        self.description = description
        self.hidden = hidden
        self.placeholder = placeholder
        self.code = code
        self.commodity_scu = commodity_scu
        if children:
            self.children[:] = children

    def object_to_validate(self, change):
        if change[-1] != "deleted":
            yield self

    def validate(self):
        if self.type not in ACCOUNT_TYPES:
            raise ValueError("Account_type '{}' is not in {}".format(self.type, ACCOUNT_TYPES))

        if self.parent:
            if not _is_parent_child_types_consistent(self.parent.type, self.type, self.book.control_mode):
                raise ValueError("Child type '{}' is not consistent with parent type {}".format(
                    self.type, self.parent.type))

            for acc in self.parent.children:
                if acc.name == self.name and acc != self:
                    raise ValueError(
                        "{} has two children with the same name {} : {} and {}".format(self.parent, self.name,
                                                                                       self, acc))
        else:
            if self.type in root_types:
                if self.name not in ['Template Root', 'Root Account']:
                    raise ValueError("{} is a root account but has a name = '{}'".format(self, self.name))
            else:
                raise ValueError("{} has no parent but is not a root account".format(self))

    @validates('commodity')
    def observe_commodity(self, key, value):
        """
        Ensure update of commodity_scu when commodity is changed
        """
        if value and (self.commodity_scu is None or self.non_std_scu == 0):
            self.commodity_scu = value.fraction

        return value

    @property
    def fullname(self):
        if self.parent:
            pfn = self.parent.fullname
            if pfn:
                return "{}:{}".format(pfn, self.name)
            else:
                return self.name
        else:
            return ""

    def get_balance(self, recurse=True, commodity=None, natural_sign=True):
        """
        Returns the balance of the account (including its children accounts if recurse=True)
        expressed in account's commodity/currency.
        If this is a stock/fund account, it will return the number of shares held.
        If this is a currency account, it will be in account's currency.
        In case of recursion, the commodity of children accounts will be transformed to the commodity of the father account using the latest price
        (if no price is available to convert , it is considered as 0).
        If natural_sign is True, the sign of the balance is reverted for the account with type {'LIABILITY', 'PAYABLE', 'CREDIT', 'INCOME', 'EQUITY'}

        Attributes:
            recurse (bool, optional): True if the balance should include children accounts (default to True)
            commodity (:class:`piecash.core.commodity.Commodity`): the currency into which to get the balance (default to None, i.e. the currency of the account)
            natural_sign (bool, optional): True if the balance sign is reversed for accounts of type {'LIABILITY', 'PAYABLE', 'CREDIT', 'INCOME', 'EQUITY'} (default to True)

        Returns:
            the balance of the account
        """
        if commodity is None:
            commodity = self.commodity

        balance = sum([sp.quantity for sp in self.splits])

        if commodity != self.commodity:
            try:
                # conversion is done directly from self.commodity to commodity (if possible)
                factor = self.commodity.currency_conversion(commodity)
                balance = balance * factor
            except GncConversionError:
                # conversion is done from self.commodity to self.parent.commodity and then to commodity
                factor1 = self.commodity.currency_conversion(self.parent.commodity)
                factor2 = self.parent.commodity.currency_conversion(commodity)
                factor = factor1 * factor2
                balance = balance * factor

        if recurse and self.children:
            balance += sum(acc.get_balance(recurse=recurse, commodity=commodity, natural_sign=False) for acc in self.children)

        if natural_sign:
            return balance * self.sign
        else:
            return balance

    @property
    def sign(self):
        return 1 if (self.type in positive_types) else -1

    @property
    def is_template(self):
        return self.commodity.namespace == 'template'

    def __str__(self):
        if self.commodity:
            return "Account<{acc.fullname}[{acc.commodity.mnemonic}]>".format(acc=self)
        else:
            return "Account<{acc.fullname}>".format(acc=self)



from piecash import open_book

# open a book
with open_book("../gnucash_books/simple_sample.gnucash", open_if_lock=True) as mybook:
    # print all splits in account "Asset"
    asset = mybook.accounts(fullname="Asset")
    for split in asset.splits:
        print(split)

    # extract all split information to a pandas DataFrame
    df = mybook.splits_df()

    # print for account "Asset" some information on the splits
    print(df.loc[df["account.fullname"] == "Asset", ["transaction.post_date", "value"]])



import json
import logging
import os
from collections import namedtuple

__author__ = "sdementen"


def quandl_fx(fx_mnemonic, base_mnemonic, start_date):
    """Retrieve exchange rate of commodity fx in function of base.

    API KEY will be retrieved from the environment variable QUANDL_API_KEY
    """
    import requests

    PUBLIC_API_URL = "https://www.quandl.com/api/v1/datasets/CURRFX/{}{}.json".format(
        fx_mnemonic, base_mnemonic
    )
    params = {
        "request_source": "python",
        "request_version": 2,
        "trim_start": "{:%Y-%m-%d}".format(start_date),
    }

    # adapt quandl parameters  with apikey if given in environment variable QUANDL_API_KEY
    apikey = os.environ.get("QUANDL_API_KEY")
    if apikey:
        params["api_key"] = apikey

    text_result = requests.get(PUBLIC_API_URL, params=params).text
    try:
        query_result = json.loads(text_result)
    except ValueError:
        logging.error("issue when retrieving info from quandl.com : '{}'".format(text_result))
        return []
    if "error" in query_result:
        logging.error(
            "issue when retrieving info from quandl.com : '{}'".format(query_result["error"])
        )
        return []
    if "quandl_error" in query_result:
        logging.error(
            "issue when retrieving info from quandl.com : '{}'".format(query_result["quandl_error"])
        )
        return []
    if "errors" in query_result and query_result["errors"]:
        logging.error(
            "issue when retrieving info from quandl.com : '{}'".format(query_result["errors"])
        )
        return []

    rows = query_result["data"]

    qdl_result = namedtuple("QUANDL", ["date", "rate", "high", "low"])

    return [qdl_result(*v) for v in rows]



# coding=utf-8
from __future__ import unicode_literals
import sys
from piecash import open_book, Transaction, Split
from piecash.kvp import Slot
# from gnucash import Session, Account, Transaction, Split, GncNumeric
# import gnucash
from piecash import create_book, Account, Split, Transaction, Commodity, Price
from datetime import datetime

bookname = "/home/sdementen/Projects/piecash/tests/foozbar.sqlite"
bookname = "/home/sdementen/Projects/piecash/gnucash_books/complex_example_piecash.gnucash"

with create_book(bookname, currency="EUR", keep_foreign_keys=False, overwrite=True) as b:
    # create some accounts
    curr = b.default_currency
    other_curr = b.currencies(mnemonic="USD")
    cdty = Commodity(namespace=u"BEL20", mnemonic=u"GnuCash Inc.", fullname=u"GnuCash Inc. stock")
    asset = Account(name="asset", type="ASSET", commodity=curr, parent=b.root_account)
    foreign_asset = Account(name="foreign asset", type="ASSET", commodity=other_curr, parent=b.root_account)
    stock = Account(name="broker", type="STOCK", commodity=cdty, parent=asset)
    expense = Account(name="exp", type="EXPENSE", commodity=curr, parent=b.root_account)
    income = Account(name="inc", type="INCOME", commodity=curr, parent=b.root_account)

    tr1 = Transaction(post_date=datetime(2015, 10, 21),
                      description="my revenue",
                      currency=curr,
                      splits=[
                          Split(account=asset, value=(1000, 1)),
                          Split(account=income, value=(-1000, 1)),
                      ]
                      )
    tr2 = Transaction(post_date=datetime(2015, 10, 25),
                      description="my expense",
                      currency=curr,
                      splits=[
                          Split(account=asset, value=(-100, 1)),
                          Split(account=expense, value=(20, 1), memo="cost of X"),
                          Split(account=expense, value=(80, 1), memo="cost of Y"),
                      ]
                      )
    tr_stock = Transaction(post_date=datetime(2015, 10, 29),
                           description="my purchase of stock",
                           currency=curr,
                           splits=[
                               Split(account=asset, value=(-200, 1)),
                               Split(account=expense, value=(15, 1), memo="transaction costs"),
                               Split(account=stock, value=(185, 1), quantity=(6, 1), memo="purchase of stock"),
                           ]
                           )
    tr_to_foreign = Transaction(post_date=datetime(2015, 10, 30),
                                description="transfer to foreign asset",
                                currency=curr,
                                splits=[
                                    Split(account=asset, value=(-200, 1)),
                                    Split(account=foreign_asset, value=(200, 1), quantity=(135, 1)),
                                ]
                                )
    tr_from_foreign = Transaction(post_date=datetime(2015, 10, 31),
                                  description="transfer from foreign asset",
                                  currency=other_curr,
                                  splits=[
                                      Split(account=asset, value=(135, 1), quantity=(215, 1)),
                                      Split(account=foreign_asset, value=(-135, 1)),
                                  ]
                                  )
    Price(commodity=cdty,
          currency=other_curr,
          date=datetime(2015, 11, 1),
          value=(123, 100),
          )
    Price(commodity=cdty,
          currency=other_curr,
          date=datetime(2015, 11, 4),
          value=(127, 100),
          )
    Price(commodity=cdty,
          currency=curr,
          date=datetime(2015, 11, 2),
          value=(234, 100),
          )

    b.save()

    print(b.prices_df().to_string())
    print(b.prices)

fdsdfsfds
with open_book(bookname, readonly=True, open_if_lock=True) as b:
    for sp in b.splits:
        print(sp.transaction.post_date, sp.transaction.enter_date, "=>", sp.slots)
    for p in b.prices:
        print(p._value_denom, p._value_num, p.slots)

    print("\n".join(map(str, b.session.query(Slot).all())))
dfdsffd

if False:
    sys.path.append("/home/sdementen/Apps/lib/python2.7/site-packages")


    def lookup_account_by_path(parent, path):
        acc = parent.lookup_by_name(path[0])
        if acc.get_instance() == None:
            raise Exception('Account path {} not found'.format(':'.join(path)))
        if len(path) > 1:
            return lookup_account_by_path(acc, path[1:])
        return acc


    def lookup_account(root, name):
        path = name.split(':')
        return lookup_account_by_path(root, path)


    session = Session("/path/to/file.gnucash")  # , ignore_lock=True)
    # or use URI string: ('mysql://USER:PASSWORD@HOST/DATABASE')

    today = datetime.now()
    book = session.book  # All actions are performed through the book object (or its children)
    root = book.get_root_account()  # Parent of all accounts
    currency = book.get_table().lookup('ISO4217', "USD")
    tx = Transaction(book)
    tx.BeginEdit()
    tx.SetCurrency(currency)
    tx.SetDateEnteredTS(today)
    tx.SetDatePostedTS(today)  # or another datetime object for the transaction's "register date"
    tx.SetDescription("Transaction Description!")
    # tx.SetNum(int_variable) # if you need a transaction number
    amount = 24
    sp1 = Split(book)  # First half of transaction
    sp1.SetParent(tx)
    # The lookup string needs to match your account path exactly.
    sp1.SetAccount(lookup_account(root, "Expenses:Some Expense Account"))
    # amount is an int (no $ or .), so $5.23 becomes amount=523
    sp1.SetValue(GncNumeric(amount, 100))  # Assuming you only have one split
    # For multiple splits, you need to make sure the totals all balance out.
    sp1.SetAmount(GncNumeric(amount, 100))
    sp1.SetMemo("Split Memo!")  # optional

    sp2 = Split(book)  # Need a balancing split
    sp2.SetParent(tx)
    sp2.SetAccount(lookup_account(root, "Assets:Current Assets:Checking"))
    sp2.SetValue(sp1.GetValue().neg())
    sp2.SetAmount(sp1.GetValue().neg())
    sp2.SetMemo("Other Split Memo!")  # optional

    tx.CommitEdit()  # Finish editing transaction
    session.save()
    session.end()

    fdsfsfds

if True:

    with create_book("../gnucash_books/example.gnucash", overwrite=True) as mybook:
        mybook.root_account.children = [
            Account(name="Expenses",
                    type="EXPENSE",
                    commodity=mybook.currencies(mnemonic="USD"),
                    placeholder=True,
                    children=[
                        Account(name="Some Expense Account",
                                type="EXPENSE",
                                commodity=mybook.currencies(mnemonic="USD")),
                    ]),
            Account(name="Assets",
                    type="ASSET",
                    commodity=mybook.currencies(mnemonic="USD"),
                    placeholder=True,
                    children=[
                        Account(name="Current Assets",
                                type="BANK",
                                commodity=mybook.currencies(mnemonic="USD"),
                                placeholder=True,
                                children=[
                                    Account(name="Checking",
                                            type="BANK",
                                            commodity=mybook.currencies(mnemonic="USD"))
                                ]),
                    ]),
        ]
        mybook.save()

    from piecash import open_book, Transaction, Split
    from datetime import datetime
    from decimal import Decimal

    with open_book("../gnucash_books/example.gnucash",
                   open_if_lock=True,
                   readonly=False) as mybook:
        today = datetime.now()
        # retrieve the currency from the book
        USD = mybook.currencies(mnemonic="USD")
        # define the amount as Decimal
        amount = Decimal("25.35")
        # retrieve accounts
        from_account = mybook.accounts(fullname="Expenses:Some Expense Account")
        to_account = mybook.accounts(fullname="Assets:Current Assets:Checking")
        # create the transaction with its two splits
        Transaction(
            post_date=today,
            enter_date=today,
            currency=USD,
            description="Transaction Description!",
            splits=[
                Split(account=to_account,
                      value=amount,
                      memo="Split Memo!"),
                Split(account=from_account,
                      value=-amount,
                      memo="Other Split Memo!"),
            ]
        )
        # save the book
        mybook.save()

    fdsfdsfds

    # create a book (in memory)
    b = open_book("../gnucash_books/book_schtx.gnucash", open_if_lock=True)
    print("============================================")
    df_splits = b.splits_df()

    # df_prices = b.prices_df()
    # assert isinstance(df_prices, pandas.DataFrame)

    fdsfdsfsd
    # get the currency
    eur = b.default_currency

    # create a customer
    c1 = Customer(name="Mickey", currency=eur, address=Address(addr1="Sesame street 1", email="mickey@example.com"))
    # the customer has not yet an ID
    b.add(c1)

    # flush the book
    b.flush()
    sps = b.splits
    print(sps)
    for sp in sps:
        print(sp.value)
    # the customer gets its ID
    print(c1)
    print(b.splits_df())
    fdsfds

    # or create a customer directly in a book (by specifying the book argument)
    c2 = Customer(name="Mickey", currency=eur, address=Address(addr1="Sesame street 1", email="mickey@example.com"),
                  book=b)

    # the customer gets immediately its ID
    c2

    # the counter of the ID is accessible as
    b.counter_customer

    fdsdsffds
from decimal import Decimal
from datetime import datetime
import decimal
import inspect
from sqlalchemy.orm import object_session
from piecash import open_book, Budget, Address
from piecash._common import Recurrence
from piecash import create_book, Account, Transaction, Split, Commodity, Vendor
from piecash.business import Customer

b = create_book("foo.sqlite", currency="EUR", keep_foreign_keys=False, overwrite=True)
c = Customer(book=b, name="foo", currency=b.currencies(mnemonic="EUR"),
             address=Address(name="foo", addr1="a1", addr4="a4", fax="fax", email="email", phone="phoen"))
print(c.addr_addr1)
b.add(Customer(name="john", id=456, currency=b.currencies(mnemonic="EUR")))
b.add(Customer(name="baz", currency=b.currencies(mnemonic="EUR")))
b.add(Customer(name="dsdsbaz", tax_included="YES", currency=b.currencies(mnemonic="EUR")))
b.add(Vendor(name="dsdsbaz", tax_included="YES", currency=b.currencies(mnemonic="EUR")))
b.save()
print(b.customers)
print(b.vendors)
fdsfdsfds
# create some accounts
curr = b.currencies[0]
cdty = Commodity(namespace="échange", mnemonic="ïoà", fullname="Example of unicode déta")
a = Account(name="asset", type="ASSET", commodity=curr, parent=b.root_account)
Account(name="broker", type="STOCK", commodity=cdty, parent=a)
Account(name="exp", type="EXPENSE", commodity=curr, parent=b.root_account)
Account(name="inc", type="INCOME", commodity=curr, parent=b.root_account)
b.flush()
EUR = b.commodities(namespace="CURRENCY")
racc = b.root_account
a = b.accounts(name="asset")
s = b.accounts(name="broker")
b.book.use_trading_accounts = True
tr = Transaction(currency=EUR, description="buy stock", notes="on St-Eugène day",
                 post_date=datetime(2014, 1, 2),
                 enter_date=datetime(2014, 1, 3),
                 splits=[
                     Split(account=a, value=100, memo="mémo asset"),
                     Split(account=s, value=-90, memo="mémo brok"),
                 ])
sp = tr.splits(account=s)
print(sp)
sp.quantity = -15
# adjust balance
Split(account=a, value=-10, memo="missing exp", transaction=tr)
b.flush()
sp.quantity = -14
b.flush()
# print (tr.splits)
sp.quantity = -12
b.flush()
# print (tr.splits)
sp.quantity = -13
b.flush()
sp.quantity = -13
b.flush()
print(tr.splits)

fdsfdsfd
# b1 = create_book("foo.gnucash",overwrite=True,echo=True)
# b1.save()
b1 = open_book("foo.gnucash", readonly=False, echo=True, do_backup=False)
b2 = open_book("foo.gnucash", readonly=False, echo=True, do_backup=False)
# a1 = Account("Acc 1", "ASSET", b1.default_currency, parent=b1.root_account)
# b1.flush()
# b1.save()
a = b2.accounts[0].name
# b2["fooo"] = b2.accounts[0].name
b2.accounts[0].name = "hello you2"
b1.accounts[0].name = "hello me1"
b1.flush()
b1.save()
b2.flush()
b2.save()
print(a, b1.accounts[0].name, b2.accounts[0].name)
fdsffds

# create a book (in memory)
from piecash.ledger import ledger

b = create_book(currency="EUR")
b.control_mode = ["allow-root-subaccounts"]
# get the EUR and create the USD currencies
c1 = b.default_currency
# create two accounts
a1 = Account("Acc 1", "ROOT", c1, parent=b.root_account)
b.add(a1)
print(object_session(a1).book)
print(a1)
print(b.currencies(mnemonic="USD"))
a2 = Account("Acc 2", "ROOT", c1, parent=a1)
print(a2)
b.save()
a1.name = "foo"
b.save()
tr = Transaction(b.default_currency,
                 splits=[
                     Split(account=a1, value=100, quantity=10),
                     Split(account=a2, value=-100, quantity=20)
                 ])
b.save()
# del tr.splits[0]
# b.save()
print(ledger(b))

fdsfds
# create a transaction from a1 to a2
tr = Transaction(currency=c1,
                 description="transfer",
                 splits=[
                     Split(account=a1, value=-100),
                     Split(account=a2, value=100, quantity=30)
                 ])
s.flush()

# ledger_str() returns a representation of the transaction in the ledger-cli format
tr.ledger_str()

# change the book to use the "trading accounts" options
s.book.use_trading_accounts = True
# add a new transaction identical to the previous
tr2 = Transaction(currency=c1,
                  description="transfer 2",
                  splits=[
                      Split(account=a1, value=-100),
                      Split(account=a2, value=100, quantity=30)
                  ])
tr2.ledger_str()
# when flushing, the trading accounts are created
s.flush()
tr2.ledger_str()

# trying to create an unbalanced transaction trigger an exception
# (there is not automatic creation of an imbalance split)
tr3 = Transaction(currency=c1,
                  description="transfer imb",
                  splits=[
                      Split(account=a1, value=-100),
                      Split(account=a2, value=100, quantity=30)
                  ])
print(tr3.ledger_str())
s.flush()

fdsfdsfds
if True:
    from piecash import create_book, Account

    with create_book(currency="EUR") as s:
        # retrieve the default currency
        EUR = s.commodities.get(mnemonic="EUR")

        # creating a placeholder account
        acc = Account(name="My account",
                      type="ASSET",
                      parent=s.book.root_account,
                      commodity=EUR,
                      placeholder=True, )

        # creating a detailed sub-account
        subacc = Account(name="My sub account",
                         type="BANK",
                         parent=acc,
                         commodity=EUR,
                         commodity_scu=1000,
                         description="my bank account",
                         code="FR013334...", )
        print(acc)
        s.save()
dfds

from piecash import create_book, Commodity

# create a book (in memory) with some currency
s = create_book(currency="EUR")

print(s.commodities)
print(s, s.book, s.book.gnc_session)

# creating a new ISO currency (if not already available in s.commodities)
USD = s.book.create_currency_from_ISO("USD")
print(USD)

# create a commodity
apple = s.book.create_stock_from_symbol("AAPL")
print(apple)

XBT = Commodity(namespace="CURRENCY", mnemonic="XBT", fullname="Bitcoin", fraction=1000000)
print(XBT)
cxwcxwcxw
# with create_book("empty.gnucash", overwrite=True) as s:
# print(list(s.book.iteritems()))

# with open_book("test_bitcoin_piecash.gnucash", readonly=False, open_if_lock=True, backup=False) as s:
# for tr in s.transactions:
# print(tr.ledger_str())
# with create_book("super_empty_piecash.gnucash") as s:
# pass
# dffdsfds


with open_book("super_empty_piecash.gnucash", readonly=False, open_if_lock=True, backup=False) as s:
    print(s.book.root_account.commodity)
    print(s.commodities)
    fdfdsfds
    sa = s.session
    sql = sa.query(Split.value).filter_by(value=100)
    print(sql)
    print(sql.all())
    # print(list(sa.execute("select * from splits")))

dsdsqdsq
#
with open_book("test_bitcoin_piecash.gnucash", readonly=False, open_if_lock=True, backup=False) as s:
    # print(s.book.use_trading_accounts)
    # print(s.book.RO_threshold_day)
    # s.book.use_split_action_field=True
    # print(list(s.book.iteritems()))
    # s.book.RO_threshold_day=0
    # s.book.use_trading_accounts=not s.book.use_trading_accounts
    # print(s.book.use_trading_accounts)
    # print(list(s.book.iteritems()))
    # s.book.RO_threshold_day=12
    # s.book.use_trading_accounts=not s.book.use_trading_accounts
    # print(s.book.use_trading_accounts)
    # print(list(s.book.iteritems()))
    # fdfdsfds
    s.book.use_trading_accounts = True

    # mtr = s.transactions(description="weird transaction")
    # print(mtr.ledger_str())
    # print(mtr.calculate_imbalances())
    # fdsfds

    for tr in s.transactions:
        if tr.description == "weird transaction":
            print(tr.ledger_str())
            tr.splits = [sp for sp in tr.splits if sp.account.type != "TRADING"]

        before = tr.ledger_str()
        if any(tr.calculate_imbalances()[1]):
            print(before)
            tr.normalize_trading_accounts()
            print(tr.ledger_str())

    s.save()
fdsfdsfds

with create_book("test_bitcoin.gnucash", overwrite=True) as s:
    root = s.book.root_account
    bitcoin = Commodity("CURRENCY", "XBT", "Bitcoin", 1000000)
    Account("My bitcoin account", "BANK", bitcoin, parent=root)
    s.save()
fdsfds

with open_book("../gnucash_books/default_book.gnucash") as s:
    # accessing the book object from the session
    book = s.book

    # accessing root accounts
    root = book.root_account

    # accessing children accounts of root
    r = s.book.root_account.children(name="Assets").children[0]
    for acc in s.book.root_account.children(name="Assets").children[0].children:
        print(acc)

    print(inspect.getmro(Budget))
    print(inspect.getmro(Recurrence))
    b = Budget(name=lambda x: x, foo="3")
    b = Budget(name=lambda x: x, foo="3")
    fdsd
    print(b)
fdsfds

with create_book() as s:
    # retrieve list of slots
    print(s.book.slots)

    # set slots
    s.book["myintkey"] = 3
    s.book["mystrkey"] = "hello"
    s.book["myboolkey"] = True
    s.book["mydatekey"] = datetime.datetime.today().date()
    s.book["mydatetimekey"] = datetime.datetime.today()
    s.book["mynumerickey"] = decimal.Decimal("12.34567")
    s.book["account"] = s.book.root_account

    # iterate over all slots
    for k, v in s.book.iteritems():
        print("slot={v} has key={k} and value={v.value} of type {t}".format(k=k, v=v, t=type(v.value)))

    # delete a slot
    del s.book["myintkey"]
    # delete all slots
    del s.book[:]

    # create a key/value in a slot frames (and create them if they do not exist)
    s.book["options/Accounts/Use trading accounts"] = "t"
    # access a slot in frame in whatever notations
    s1 = s.book["options/Accounts/Use trading accounts"]
    s2 = s.book["options"]["Accounts/Use trading accounts"]
    s3 = s.book["options/Accounts"]["Use trading accounts"]
    s4 = s.book["options"]["Accounts"]["Use trading accounts"]
    assert s1 == s2 == s3 == s4

dsqdsq
with open_book("/home/sdementen/Desktop/test_sch_txn_sqlite.gnucash", acquire_lock=False, open_if_lock=True) as s:
    print(s.book.root_template.children)
    print(s.commodities)
    for tr in s.transactions:
        print(tr.splits[0].slots)
        print(tr.slots)
    print(s.transactions[0].scheduled_transaction)
    s.transactions[0].scheduled_transaction = None
    # print(s.transactions[0]["from-sched-xaction"])
    s.book["a/n"]
    print(s.accounts.get(name='Checking Account').lots.get(title="Lot 3"))
fdsfdssfd

dsqdsqdsq
with open_book("../gnucash_books/simple_sample.gnucash", acquire_lock=False, open_if_lock=True) as s:
    asset = s.accounts.get(name="Asset")
    expense = s.accounts.get(name="Expense")
    eur = asset.commodity
    tr = Transaction(currency=eur,
                     description="test",
                     splits=[Split(asset, 100), Split(expense, -100)])
    print(tr)
dffdsfsd

from piecash import open_book, create_book, Account, Transaction, Split

# with create_book("all_account_types.gnucash", overwrite=True) as s:
# for actype in ACCOUNT_TYPES:
# if actype=="ROOT":
# continue
#         acc = Account(name=actype, account_type=actype, parent=s.book.root_account, commodity=s.book.root_account.commodity)
#     s.save()

with open_book("simple_sample.gnucash", acquire_lock=False, open_if_lock=True) as s:
    asset = s.accounts.get(name="Asset")
    expense = s.accounts.get(name="Expense")
    eur = asset.commodity
    tr = Transaction(currency=eur,
                     description="test",
                     splits=[Split(asset, 100), Split(expense, -100)])
    print(tr)
    fdsdf
    for tr in s.transactions:
        for sp in tr.splits:
            sp.account = asset

    fdsfsdfd

    for tr in s.transactions:
        print(tr)
        for sp in tr.splits:
            print("\t", sp)

    for acc in s.accounts:
        if acc.commodity:
            print(acc.commodity.base_currency)
            bal = acc.get_balance()
            print("{} : {} {}".format(acc.fullname, bal, acc.commodity.mnemonic))

fsdsfsd

# with create_book("test_simple_transaction.gnucash",overwrite=True) as s:
with create_book(currency="EUR") as s:
    EUR = s.commodities.get(mnemonic="EUR")
    acc1 = Account(name="acc1",
                   commodity=EUR,
                   parent=s.book.root_account,
                   account_type="BANK")
    acc2 = Account(name="acc2",
                   commodity=EUR,
                   parent=s.book.root_account,
                   account_type="BANK",
                   commodity_scu=10)
    s.session.flush()

    tr = Transaction(currency=EUR,
                     description="foo",
                     splits=[
                         Split(value=Decimal("1.2345"),
                               account=acc1),
                         Split(value=Decimal("-1.2345"),
                               account=acc2),
                     ])
    s.session.flush()

    print([(sp._quantity_denom, sp._value_denom) for sp in tr.splits])
    print(s.session.query(Account.commodity).all())
    print(tr.slots)
    tr.post_date = datetime.datetime.now()
    print(tr.slots)
    tr.post_date = None
    print(tr.slots)
    print(acc2.slots)
    acc2.placeholder = 1
    print(acc2.slots)
    acc2.placeholder = 0
    print(acc2.slots)
    s.save()
    # del tr.splits[-1]
    # print tr.get_imbalances()

dsqsdqdqs
#
# with create_book("test.gnucash", keep_foreign_keys=False, overwrite=True) as s:
# EUR = Commodity.create_from_ISO("EUR")
# tr = Transaction(currency=EUR,description="first")
#     sp = Split(transaction=tr, account=s.book.root_account, value=(100,1), quantity=(10,1))
#     tr = Transaction(currency=EUR,description="second")
#     s.save()


# with open_book("trading_accounts.gnucash", readonly=False) as s:
#     for tr in s.transactions:#.get(description="other transfer + expense")
#         print "{}\t{}".format(tr.currency, tr.description)
#         # print tr.get_imbalances()
#         for sp in tr.splits:
#             print "\t[{}] {} / {} for {}".format( sp.account.commodity,sp.value, sp.quantity,sp.account)
#     # tr1 = s.transactions.get(description="first")
#     # tr2 = s.transactions[1]
#     tr2 = s.transactions.get(description="cross CAD to USD transfer (initiated from USD account)")
#     sp = s.transactions.get(description="cross CAD to USD transfer").splits[0]
#     # sp = s.transactions[0].splits[0]
#     print "o"*100
#     # print sp
#     print "o"*100
#     sp.transaction = tr2
#     s.session.flush()
# fdfdsfsd

with open_book("trading_accounts.gnucash", readonly=False, open_if_lock=True, acquire_lock=True) as s:
    for tr in s.transactions:  # .get(description="other transfer + expense")
        print("{}\t{}".format(tr.currency, tr.description))
        # print tr.get_imbalances()
        for sp in tr.splits:
            print("\t[{}] {} / {} for {}".format(sp.account.commodity, sp.value, sp.quantity, sp.account))
    # sp.memo = "foo"
    # tr.description = "foo"
    # s.session.flush()

    tr = s.transactions.get(description="cross CAD to USD transfer (initiated from USD account)")
    sp = s.transactions.get(description="cross CAD to USD transfer").splits[0]
    # sp.transaction = tr
    tr.description = "foo"
    # tr.currency = s.commodities[-1]
    # print "foooooooooooooooooooooooooooooooo"
    # print sp.transaction, s.transactions.get(description="cross CAD to USD transfer").splits[1].transaction

    acc = s.accounts[0]
    assert isinstance(acc, Account)
    acc.commodity_scu = 1

    s.session.flush()
    Transaction(currency=s.commodities.get(mnemonic="EUR"),
                description="foo",
                splits=[
                    Split(value=Decimal("1.2345"),
                          account=s.accounts[0]),
                    Split(value=Decimal("1.2345"),
                          account=s.accounts[1]),
                ])
    s.session.flush()
    # del tr.splits[-1]
    # print tr.get_imbalances()
fdfsdfds

# s = create_book(postgres_conn="postgres://user:passwd@localhost/gnucash_book1", overwrite=True)
# s = create_book("test.gnucash",overwrite=True)

s = create_book("test.gnucash", overwrite=True)
# s = create_book()
EUR = Commodity.create_from_ISO("EUR")
CAD = Commodity.create_from_ISO("CAD")
USD = Commodity.create_from_ISO("USD")
# EUR.fraction = 100000
acc1 = Account(name="foo EUR", parent=s.book.root_account, account_type="ASSET", placeholder=False, commodity=EUR)
acc2 = Account(name="baz CAD", parent=s.book.root_account, account_type="STOCK", placeholder=False, commodity=CAD)
acc3 = Account(name="baz USD", parent=s.book.root_account, account_type="STOCK", placeholder=False, commodity=USD)
# acc1.commodity_scu = 1000
t = Transaction(description="foo",
                post_date=datetime.datetime.now(),
                enter_date=datetime.datetime.now(),
                currency=EUR)
Split(transaction=t,
      value=-25,
      quantity=-25,
      account=acc1)
Split(transaction=t,
      value=15,
      quantity=10,
      account=acc2)
Split(transaction=t,
      value=11,
      quantity=12,
      account=acc3)
print(t.splits)
imb = t.get_imbalances()
print(imb)
t.add_imbalance_splits()
s.save()
# print s.session.query(Split).filter(Split.value_magic>=1.0).all()
ffdsfdd

# print t.splits[0].value_gnc, t.splits[0].value_denom, t.splits[0].value_num

print(s.book.root_account.children.append(Account(name="rool", account_type="ASSET", placeholder=False, commodity=EUR)))
s.save()
print(s.transactions.get)
print(s.accounts.get)
s.close()
# end_of_example


with open_book("sample1.gnucash", readonly=False, open_if_lock=True) as s1, open_book("sample2.gnucash") as s2:
    b1 = s1.book
    # s2 = open_book("sample2.gnucash")
    b2 = s2.book

    eur = s1.get(Commodity, mnemonic="EUR")

    b1.root_account["notes"] = "Hello world!"

    p = Price(currency=eur,
              commodity=eur,
              value=Decimal("4234.342"),
              )
    s1.session.add(p)
    print(p.value)
    print(p.value_denom)
    print(p.value_num)

    with b1:
        print(get_active_session())
        with b2:
            print(get_active_session())
        print(get_active_session())

    with b1:
        acc = Account(name="foo")
    print(s1.session.new)



# -*- coding: latin-1 -*-
import glob
import os
from decimal import Decimal
from pathlib import Path

import pytest
from sqlalchemy import create_engine
from sqlalchemy.engine.reflection import Inspector
from sqlalchemy.exc import OperationalError
from sqlalchemy.orm import Session

from piecash import create_book, Account, GnucashException, Book, open_book, Commodity
from piecash.core import Version
from test_helper import (
    db_sqlite_uri,
    db_sqlite,
    new_book,
    new_book_USD,
    book_uri,
    book_transactions,
    book_sample,
    book_investment,
    book_reference_3_0_0_fulloptions,
    book_reference_3_0_0_basic,
)

# dummy line to avoid removing unused symbols
a = (
    db_sqlite_uri,
    db_sqlite,
    new_book,
    new_book_USD,
    book_uri,
    book_transactions,
    book_sample,
    book_investment,
    book_reference_3_0_0_fulloptions,
    book_reference_3_0_0_basic,
)


class TestBook_create_book(object):
    def test_create_default(self, new_book):
        assert isinstance(new_book, Book)
        assert isinstance(new_book.session, Session)
        assert new_book.uri is not None
        assert new_book.session.bind.name in ["sqlite", "postgresql", "mysql"]
        assert (
            repr(new_book.query(Version).filter_by(table_name="commodities").one())
            == "Version<commodities=1>"
        )

        EUR = new_book.commodities[0]
        assert EUR.mnemonic == "EUR"
        assert EUR.namespace == "CURRENCY"

        # no std account
        assert len(new_book.accounts) == 0
        # two root accounts
        root_accs = new_book.query(Account).all()
        assert all([acc.type == "ROOT" for acc in root_accs])
        assert len(root_accs) == 2

        # no slots
        assert len(new_book.slots) == 0

    def test_create_save_cancel_flush(self, new_book):
        EUR = new_book.commodities[0]
        EUR.mnemonic = "foo"
        assert EUR.mnemonic == "foo"
        new_book.cancel()
        assert EUR.mnemonic == "EUR"

        EUR.mnemonic = "foo"
        assert EUR.mnemonic == "foo"
        new_book.flush()
        assert EUR.mnemonic == "foo"
        new_book.cancel()
        assert EUR.mnemonic == "EUR"

        EUR.mnemonic = "foo"
        new_book.save()
        assert EUR.mnemonic == "foo"

    def test_create_USD_book(self, new_book_USD):
        CUR = new_book_USD.commodities[0]
        assert CUR.mnemonic == "USD"
        assert CUR.namespace == "CURRENCY"

    def test_create_specific_currency(self):
        b = create_book(currency="USD")
        CUR = b.commodities[0]
        assert CUR.mnemonic == "USD"
        assert CUR.namespace == "CURRENCY"

        b = create_book(currency="CHF")
        CUR = b.commodities[0]
        assert CUR.mnemonic == "CHF"
        assert CUR.namespace == "CURRENCY"

        with pytest.raises(ValueError):
            b = create_book(currency="ZIE")

    def test_create_named_sqlite_book(self):
        # remove file if left from previous test
        assert isinstance(db_sqlite, Path)
        if db_sqlite.exists():
            db_sqlite.unlink()

        # assert error if both sqlite_file and uri_conn are defined
        with pytest.raises(ValueError):
            b = create_book(db_sqlite, db_sqlite_uri)

        # assert creation of file
        b = create_book(db_sqlite)
        assert db_sqlite.exists()
        t = db_sqlite.stat().st_mtime

        # ensure error if no overwrite
        with pytest.raises(GnucashException):
            b = create_book(db_sqlite)
        assert db_sqlite.stat().st_mtime == t
        with pytest.raises(GnucashException):
            b = create_book(uri_conn="sqlite:///{}".format(db_sqlite))
        assert db_sqlite.stat().st_mtime == t
        with pytest.raises(GnucashException):
            b = create_book(db_sqlite, overwrite=False)
        assert db_sqlite.stat().st_mtime == t

        # if overwrite, DB is recreated
        b = create_book(db_sqlite, overwrite=True)
        assert db_sqlite.stat().st_mtime > t

        # clean test
        db_sqlite.unlink()

    def test_create_with_FK(self):
        # create and keep FK
        b = create_book(uri_conn=db_sqlite_uri, keep_foreign_keys=True, overwrite=True)
        b.session.close()

        insp = Inspector.from_engine(create_engine(db_sqlite_uri))
        fk_total = []
        for tbl in insp.get_table_names():
            fk_total.append(insp.get_foreign_keys(tbl))
        assert len(fk_total) == 25

    def test_create_without_FK(self):
        # create without FK
        b = create_book(uri_conn=db_sqlite_uri, keep_foreign_keys=False, overwrite=True)
        b.session.close()

        insp = Inspector.from_engine(create_engine(db_sqlite_uri))
        for tbl in insp.get_table_names():
            fk = insp.get_foreign_keys(tbl)
            assert len(fk) == 0


class TestBook_open_book(object):
    def test_open_noarg(self):
        with pytest.raises(ValueError):
            open_book()

    def test_open_default(self, book_uri):
        # open book that does not exists
        with pytest.raises(GnucashException):
            b = open_book(uri_conn=book_uri)

        # create book
        with create_book(uri_conn=book_uri):
            pass

        # assert error if both sqlite_file and uri_conn are defined on open_book
        with pytest.raises(ValueError):
            b = open_book(db_sqlite, db_sqlite_uri)

        # open book that exists
        with open_book(uri_conn=book_uri) as b:
            # try to save (raise as RO per default)
            with pytest.raises(GnucashException):
                b.save()

            # read default currency (to check reading)
            assert b.default_currency.mnemonic == "EUR"

        # open book with checking existence
        book_uri_fail = book_uri.replace("foo", "foofail")
        with pytest.raises(GnucashException, match="Database .* does not exist"):
            open_book(uri_conn=book_uri_fail)

        # open book without checking existence
        with pytest.raises(OperationalError):
            open_book(uri_conn=book_uri_fail, check_exists=False)

    def test_open_RW_backup(self, book_uri):
        # create book
        with create_book(uri_conn=book_uri) as b:
            engine_type = b.session.bind.name

        # open book with readonly = False (ie RW)
        if engine_type != "sqlite":
            # raise an exception as try to do a backup on postgres which is not supported yet
            with pytest.raises(GnucashException):
                b = open_book(uri_conn=book_uri, readonly=False)

        elif engine_type == "sqlite":
            # delete all potential existing backup files
            url = book_uri[len("sqlite:///") :]
            for fn in glob.glob("{}.[0-9]*.gnucash".format(url)):
                os.remove(fn)

            # open file in RW without a backup creation
            with open_book(uri_conn=book_uri, readonly=False, do_backup=False) as b:
                pass

            # check no backup file creation
            assert len(glob.glob("{}.[0-9]*.gnucash".format(url))) == 0

            # open file in RW without a backup creation
            with open_book(uri_conn=book_uri, readonly=False) as b:
                pass

            # check backup file creation
            assert len(glob.glob("{}.[0-9]*.gnucash".format(url))) == 1

    def test_open_lock(self, book_uri):
        # create book and set a lock
        with create_book(uri_conn=book_uri) as b:
            b.session.create_lock()
            b.save()

        # try to open locked book
        with pytest.raises(GnucashException):
            b = open_book(uri_conn=book_uri)

        # open book specifying open_if_lock as True
        with open_book(uri_conn=book_uri, open_if_lock=True) as b:
            pass

        # open book specifying open_if_lock as True and RW to delete lock
        with open_book(uri_conn=book_uri, open_if_lock=True, readonly=False, do_backup=False) as b:
            b.session.delete_lock()
            b.save()

        # open book specifying open_if_lock as False as lock has been removed
        with open_book(uri_conn=book_uri, open_if_lock=False) as b:
            pass

    def test_read_book_transactions(self, book_sample):
        assert len(book_sample.transactions) == 5


class TestBook_access_book(object):
    def test_book_options(self, new_book):
        assert new_book.use_trading_accounts == False
        assert new_book.use_split_action_field == False
        assert new_book.RO_threshold_day == 0

        new_book.use_trading_accounts = False
        new_book.use_split_action_field = False
        new_book.RO_threshold_day = 0

        assert new_book.use_trading_accounts == False
        assert new_book.use_split_action_field == False
        assert new_book.RO_threshold_day == 0

        assert len(new_book.slots) == 0

        with pytest.raises(KeyError):
            new_book["options"]

        new_book.use_trading_accounts = True
        assert new_book["options"].value == {"Accounts": {"Use Trading Accounts": "t"}}

        new_book.use_split_action_field = True
        assert new_book["options"].value == {
            "Accounts": {"Use Split Action Field for Number": "t", "Use Trading Accounts": "t"}
        }

        new_book.RO_threshold_day = 50
        assert new_book["options"].value == {
            "Accounts": {
                "Day Threshold for Read-Only Transactions (red line)": 50.0,
                "Use Split Action Field for Number": "t",
                "Use Trading Accounts": "t",
            }
        }

        new_book.RO_threshold_day = 0
        assert new_book["options"].value == {
            "Accounts": {"Use Split Action Field for Number": "t", "Use Trading Accounts": "t"}
        }

        new_book.use_split_action_field = False
        assert new_book["options"].value == {"Accounts": {"Use Trading Accounts": "t"}}

        del new_book["options"]
        with pytest.raises(KeyError):
            new_book["options"]

        assert new_book.control_mode == []

    def test_book_trading_accounts(self, new_book):
        assert len(new_book.accounts) == 0
        assert len(new_book.currencies) == 1

        # get (and create on the fly) the trading account for the default currency
        ncur = new_book.currencies(mnemonic="USD")
        cur = new_book.default_currency

        ta = new_book.trading_account(ncur)
        new_book.flush()
        assert len(new_book.currencies) == 2
        assert len(new_book.accounts) == 3
        acc = new_book.root_account.children[0]
        assert acc.name == "Trading"
        assert acc.commodity == cur
        acc = acc.children[0]
        assert acc.name == cur.namespace
        assert acc.commodity == cur
        acc = acc.children[0]
        assert acc.name == ncur.mnemonic
        assert acc.commodity == ncur

        ncur = new_book.currencies(mnemonic="CAD")
        ta = new_book.trading_account(ncur)
        new_book.flush()
        assert len(new_book.accounts) == 4
        assert len(new_book.currencies) == 3

        ncur = new_book.currencies(mnemonic="USD")
        ta = new_book.trading_account(ncur)
        assert len(new_book.accounts) == 4
        assert len(new_book.currencies) == 3

    def test_book_transactions(self, new_book):
        ncur = new_book.currencies(mnemonic="CAD")
        new_book.flush()
        assert len(new_book.currencies) == 2
        assert not new_book.is_saved
        new_book.cancel()
        assert new_book.is_saved
        assert len(new_book.currencies) == 1
        nncur = new_book.currencies(mnemonic="USD")
        new_book.flush()
        assert not new_book.is_saved
        assert len(new_book.currencies) == 2
        new_book.save()
        assert new_book.is_saved
        assert len(new_book.currencies) == 2
        new_book.delete(nncur)
        assert not new_book.is_saved
        assert len(new_book.currencies) == 2
        new_book.save()
        assert new_book.is_saved
        assert len(new_book.currencies) == 1
        nncur = new_book.currencies(mnemonic="USD")
        new_book.flush()
        assert len(new_book.currencies) == 2
        assert not new_book.is_saved

    def test_book_getters(self, new_book):
        cur = new_book.currencies[0]
        assert cur == new_book.get(Commodity, mnemonic=cur.mnemonic)

        with pytest.raises(ValueError):
            new_book.get(Commodity, mnemonic="FOO")

        with pytest.raises(ValueError):
            new_book.get(Commodity, mnemonic="CAD")

        assert new_book.get(Commodity).all() == [cur]

        assert new_book.accounts == []
        assert new_book.transactions == []
        assert new_book.commodities == [cur]
        assert new_book.currencies == [cur]
        assert new_book.prices == []
        assert new_book.customers == []
        assert new_book.vendors == []
        assert new_book.employees == []
        assert new_book.taxtables == []
        assert new_book.invoices == []

    def test_splits_df(self, book_transactions):
        df = book_transactions.splits_df().reset_index()

        # remove guid columns as not comparable from run to run
        for col in df.columns:
            if "guid" in col:
                del df[col]

        # converte datetime to date as different tzone in CI environments
        # df["transaction.post_date"] = df["transaction.post_date"].dt.date

        df_to_string = """    value quantity               memo      transaction.description transaction.post_date transaction.currency.mnemonic account.fullname account.commodity.mnemonic
0   -1000    -1000                                      my revenue            2015-10-21                           EUR              inc                        EUR
1    1000     1000                                      my revenue            2015-10-21                           EUR            asset                        EUR
2    -100     -100                                      my expense            2015-10-25                           EUR            asset                        EUR
3      20       20          cost of X                   my expense            2015-10-25                           EUR              exp                        EUR
4      80       80          cost of Y                   my expense            2015-10-25                           EUR              exp                        EUR
5    -200     -200                            my purchase of stock            2015-10-29                           EUR            asset                        EUR
6      15       15  transaction costs         my purchase of stock            2015-10-29                           EUR              exp                        EUR
7     185        6  purchase of stock         my purchase of stock            2015-10-29                           EUR     asset:broker               GnuCash Inc.
8    -200     -200                       transfer to foreign asset            2015-10-30                           EUR            asset                        EUR
9     200      135                       transfer to foreign asset            2015-10-30                           EUR    foreign asset                        USD
10   -135     -135                     transfer from foreign asset            2015-10-31                           USD    foreign asset                        USD
11    135      215                     transfer from foreign asset            2015-10-31                           USD            asset                        EUR"""

        assert df_to_string == df.to_string()

    def test_splits_df_with_additional(self, book_transactions):
        # Adding in two additional memo fields here. If it works for non-unique it should
        # be fine for unique fields.
        df = book_transactions.splits_df(additional_fields=["memo", "memo"]).reset_index()

        # remove guid columns as not comparable from run to run
        for col in df.columns:
            if "guid" in col:
                del df[col]

        # converte datetime to date as different tzone in CI environments
        # df["transaction.post_date"] = df["transaction.post_date"].dt.date

        df_to_string = """    value quantity               memo      transaction.description transaction.post_date transaction.currency.mnemonic account.fullname account.commodity.mnemonic               memo               memo
0   -1000    -1000                                      my revenue            2015-10-21                           EUR              inc                        EUR                                      
1    1000     1000                                      my revenue            2015-10-21                           EUR            asset                        EUR                                      
2    -100     -100                                      my expense            2015-10-25                           EUR            asset                        EUR                                      
3      20       20          cost of X                   my expense            2015-10-25                           EUR              exp                        EUR          cost of X          cost of X
4      80       80          cost of Y                   my expense            2015-10-25                           EUR              exp                        EUR          cost of Y          cost of Y
5    -200     -200                            my purchase of stock            2015-10-29                           EUR            asset                        EUR                                      
6      15       15  transaction costs         my purchase of stock            2015-10-29                           EUR              exp                        EUR  transaction costs  transaction costs
7     185        6  purchase of stock         my purchase of stock            2015-10-29                           EUR     asset:broker               GnuCash Inc.  purchase of stock  purchase of stock
8    -200     -200                       transfer to foreign asset            2015-10-30                           EUR            asset                        EUR                                      
9     200      135                       transfer to foreign asset            2015-10-30                           EUR    foreign asset                        USD                                      
10   -135     -135                     transfer from foreign asset            2015-10-31                           USD    foreign asset                        USD                                      
11    135      215                     transfer from foreign asset            2015-10-31                           USD            asset                        EUR                                      """

        assert df_to_string == df.to_string()

    def test_prices_df(self, book_transactions):
        df = book_transactions.prices_df().reset_index()

        # remove guid columns as not comparable from run to run
        for col in df.columns:
            if "guid" in col:
                del df[col]

        # converte datetime to date as different tzone in CI environments
        # df["date"] = df["date"].dt.date

        df_to_string = """   index        date         type      value commodity.mnemonic currency.mnemonic
0      0  2015-10-31  transaction   0.627907                EUR               USD
1      1  2015-10-29  transaction  30.833333       GnuCash Inc.               EUR
2      2  2015-11-01      unknown       1.23       GnuCash Inc.               USD
3      3  2015-11-02      unknown       2.34       GnuCash Inc.               EUR
4      4  2015-11-04      unknown       1.27       GnuCash Inc.               USD
5      5  2015-10-30  transaction   1.481481                USD               EUR"""

        assert df_to_string == df.to_string()

    def test_commodity_quantity(self, book_investment):
        """
        Tests listing the commodity quantity in the account.
        """
        security = book_investment.get(Commodity, mnemonic="VEUR")

        total = Decimal(0)

        for account in security.accounts:
            # exclude Trading accouns explicitly.
            if account.type == "TRADING":
                continue

            balance = account.get_balance()

            # print(account.fullname, balance)
            total += balance

        # print("Balance:", total_balance)
        assert total == Decimal(13)

    def test_business_slots_options(self, book_reference_3_0_0_fulloptions):
        """
        Tests business slots
        :type book_reference_2_6_21_fulloptions: Book
        """
        assert (
            book_reference_3_0_0_fulloptions.business_company_address
            == "Rue de la Chenille éclairée, 22"
        )
        assert book_reference_3_0_0_fulloptions.business_company_contact == "John Michu"
        assert book_reference_3_0_0_fulloptions.business_company_email == "woozie@example.com"
        assert book_reference_3_0_0_fulloptions.business_company_ID == "SIREN 123 456 789"
        assert book_reference_3_0_0_fulloptions.business_company_name == "Woozie Inc"
        assert book_reference_3_0_0_fulloptions.business_company_phone == "+33 1 33 33 33 33"
        assert book_reference_3_0_0_fulloptions.business_company_website == "www.woozie.com"

    def test_business_slots_nooptions(self, book_reference_3_0_0_basic):
        """
        Tests business slots
        :type book_reference_3_0_0_basic: Book
        """
        assert book_reference_3_0_0_basic.business_company_address == ""
        assert book_reference_3_0_0_basic.business_company_contact == ""
        assert book_reference_3_0_0_basic.business_company_email == ""
        assert book_reference_3_0_0_basic.business_company_ID == ""
        assert book_reference_3_0_0_basic.business_company_name == ""
        assert book_reference_3_0_0_basic.business_company_phone == ""
        assert book_reference_3_0_0_basic.business_company_website == ""

    def test_business_writeslots_nooptions(self, book_reference_3_0_0_basic):
        """
        Tests business slots
        :type book_reference_3_0_0_basic: Book
        """
        book_reference_3_0_0_basic.business_company_address = "é"
        book_reference_3_0_0_basic.business_company_contact = "à"
        book_reference_3_0_0_basic.business_company_email = "ù"
        book_reference_3_0_0_basic.business_company_ID = "ö"
        book_reference_3_0_0_basic.business_company_name = "µ"
        book_reference_3_0_0_basic.business_company_phone = "²"
        book_reference_3_0_0_basic.business_company_website = "³"

        assert book_reference_3_0_0_basic.business_company_address == "é"
        assert book_reference_3_0_0_basic.business_company_contact == "à"
        assert book_reference_3_0_0_basic.business_company_email == "ù"
        assert book_reference_3_0_0_basic.business_company_ID == "ö"
        assert book_reference_3_0_0_basic.business_company_name == "µ"
        assert book_reference_3_0_0_basic.business_company_phone == "²"
        assert book_reference_3_0_0_basic.business_company_website == "³"



#!/usr/local/bin/python
"""original script from https://github.com/MatzeB/pygnucash/blob/master/gnucash2ledger.py by Matthias Braun matze@braunis.de
 adapted for:
 - python 3 support
 - new string formatting
"""

import click

import piecash
from piecash.scripts.cli import cli


@cli.command()
@click.argument("book", type=click.Path(exists=True))
@click.option("--locale/--no-locale", default=False, help="Export currency amounts using locale for currencies format")
@click.option(
    "--commodity-notes/--no-commodity-notes",
    default=True,
    help="Include the commodity_notes for the commodity (hledger does not support commodity commodity_notes",
)
@click.option(
    "--short-account-names/--no-short-account-names",
    default=False,
    help="Use the short name for the accounts instead of the full hierarchical name.",
)
@click.option(
    "--output",
    type=click.File("w", encoding="UTF-8"),
    default="-",
    help="File to which to export the data (default=stdout)",
)
def ledger(book, output, locale, commodity_notes, short_account_names):
    """Export to ledger-cli format.

    This scripts export a GnuCash BOOK to the ledget-cli format.
    """
    with piecash.open_book(book, open_if_lock=True) as data:
        output.write(piecash.ledger(data, locale=locale, commodity_notes=commodity_notes,short_account_names=short_account_names))



import uuid

from sqlalchemy import Column, VARCHAR, event
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.orm import relation, foreign, object_session

from ._common import CallableList
from .kvp import DictWrapper, Slot
from .sa_extra import DeclarativeBase


class DeclarativeBaseGuid(DictWrapper, DeclarativeBase):
    __abstract__ = True

    #: the unique identifier of the object
    guid = Column('guid', VARCHAR(length=32), primary_key=True, nullable=False, default=lambda: uuid.uuid4().hex)

    @declared_attr
    def slots(cls):
        rel = relation('Slot',
                       primaryjoin=foreign(Slot.obj_guid) == cls.guid,
                       cascade='all, delete-orphan',
                       collection_class=CallableList,
                       )

        return rel

    # set the relation to the slots table (KVP)
    @classmethod
    def __declare_last__(cls):
        # do not do it on the DeclarativeBaseGuid as it is an abstract class
        if cls == DeclarativeBaseGuid:
            return

        # assign id of slot when associating to object
        @event.listens_for(cls.slots, "remove")
        def my_append_listener_slots(target, value, initiator):
            s = object_session(value)
            if s:
                if value in s.new:
                    s.expunge(value)
                else:
                    s.delete(value)



# coding=utf-8
from __future__ import unicode_literals

from .._common import GnucashException
from ..yahoo_client import get_latest_quote


def create_stock_accounts(cdty, broker_account, income_account=None, income_account_types="D/CL/I"):
    """Create the multiple accounts used to track a single stock, ie:

    - broker_account/stock.mnemonic

    and the following accounts depending on the income_account_types argument

    - D = Income/Dividend Income/stock.mnemonic
    - CL = Income/Cap Gain (Long)/stock.mnemonic
    - CS = Income/Cap Gain (Short)/stock.mnemonic
    - I = Income/Interest Income/stock.mnemonic

    Args:
        broker_account (:class:`piecash.core.account.Account`): the broker account where the account holding
        the stock is to be created
        income_account (:class:`piecash.core.account.Account`): the income account where the accounts holding
        the income related to the stock are to be created
        income_account_types (str): "/" separated codes to drive the creation of income accounts

    Returns:
        :class:`piecash.core.account.Account`: a tuple with the account under the broker_account where the stock is held
        and the list of income accounts.
    """
    if cdty.namespace == "CURRENCY":
        raise GnucashException("{} is a currency ! You can't create stock_accounts for currencies".format(cdty))

    from .account import Account

    symbol = cdty.mnemonic
    try:
        acc = broker_account.children(name=symbol)
    except KeyError:
        acc = Account(symbol, "STOCK", cdty, broker_account)

    inc_accounts = []
    if income_account:
        cur = cdty.base_currency

        for inc_acc in income_account_types.split("/"):
            sub_account_name = {
                "D": "Dividend Income",
                "CL": "Cap Gain (Long)",
                "CS": "Cap Gain (Short)",
                "I": "Interest Income",
            }[inc_acc]
            try:
                sub_acc = income_account.children(name=sub_account_name)
            except KeyError:
                sub_acc = Account(sub_account_name, "INCOME", cur.base_currency, income_account)
            try:
                cdty_acc = sub_acc.children(name=symbol)
            except KeyError:
                cdty_acc = Account(symbol, "INCOME", cur, sub_acc)
            inc_accounts.append(cdty_acc)

    return acc, inc_accounts


def create_currency_from_ISO(isocode):
    """
    Factory function to create a new currency from its ISO code

    Args:
        isocode (str): the ISO code of the currency (e.g. EUR for the euro)

    Returns:
        :class:`Commodity`: the currency as a commodity object
    """
    from .commodity import Commodity

    # if self.get_session().query(Commodity).filter_by(isocode=isocode).first():
    #     raise GncCommodityError("Currency '{}' already exists".format(isocode))

    from .currency_ISO import ISO_currencies

    cur = ISO_currencies.get(isocode)

    if cur is None:
        raise ValueError("Could not find the ISO code '{}' in the ISO table".format(isocode))

    # create the currency
    cdty = Commodity(mnemonic=cur.mnemonic,
                     fullname=cur.currency,
                     fraction=10 ** int(cur.fraction),
                     cusip=cur.cusip,
                     namespace="CURRENCY",
                     quote_flag=1,
                     )

    # self.gnc_session.add(cdty)
    return cdty


def create_stock_from_symbol(symbol, book=None):
    """
    Factory function to create a new stock from its symbol. The ISO code of the quoted currency of the stock is
    stored in the slot "quoted_currency".

    Args:
        symbol (str): the symbol for the stock (e.g. YHOO for the Yahoo! stock)

    Returns:
        :class:`Commodity`: the stock as a commodity object

    .. note::
       The information is gathered from the yahoo-finance package
       The default currency in which the quote is traded is stored in a slot 'quoted_currency'

    .. todo::
       use 'select * from yahoo.finance.sectors' and 'select * from yahoo.finance.industry where id ="sector_id"'
       to retrieve name of stocks and allow therefore the creation of a stock by giving its "stock name" (or part of it).
       This could also be used to retrieve all symbols related to the same company
    """
    from .commodity import Commodity

    share = get_latest_quote(symbol)

    stock = Commodity(mnemonic=symbol,
                      fullname=share.name,
                      fraction=10000,
                      namespace=share.exchange,
                      quote_flag=1,
                      quote_source="yahoo",
                      quote_tz=share.timezone,
                      )

    if book:
        book.add(stock)
        book.flush()

    return stock


def single_transaction(post_date,
                       enter_date,
                       description,
                       value,
                       from_account,
                       to_account):
    from . import Transaction, Split
    # currency is derived from "from_account" (as in GUI)
    currency = from_account.commodity
    # currency of other destination account should be identical (as only one value given)
    assert currency == to_account.commodity, "Commodities of accounts should be the same"
    tx = Transaction(
        currency=currency,
        post_date=post_date,
        enter_date=enter_date,
        description=description,
        splits=[
            Split(account=from_account, value=-value),
            Split(account=to_account, value=value),
        ])
    return tx



from piecash import open_book, ledger, Split

# create a book with some account tree structure
with open_book(
        "../gnucash_books/simple_book_transaction_creation.gnucash", readonly=True
) as mybook:
    # iterate on all the transactions in the book
    for transaction in mybook.transactions:
        # add some extra text to the description
        transaction.description = transaction.description + " (some extra info)"
        # iterate over all the splits
        # as we will add splits to the transaction in the loop,
        # we need to use list(...) to make a list of the splits at the start of the loop
        for split in list(transaction.splits):
            # create the new split (here a copy of the each existing split
            # in the transaction with value/quantity divided by 10
            new_split = Split(
                account=split.account,
                value=split.value / 10,
                quantity=split.quantity / 10,
                memo="my new split",
                transaction=transaction,  # attach the split to the current transaction
            )
    mybook.flush()  # register the changes (but not save)
    print(ledger(mybook))

    # save the book
    # this will raise an error as readonly=True (change to readonly=False to successfully save the book)
    mybook.save()



import uuid
from sqlalchemy import Column, VARCHAR, BIGINT, INTEGER, ForeignKey
from sqlalchemy.orm import relation

from .._common import hybrid_property_gncnumeric, CallableList
from .._declbase import DeclarativeBaseGuid, DeclarativeBase
from ..sa_extra import ChoiceType


class Taxtable(DeclarativeBaseGuid):
    __tablename__ = 'taxtables'

    __table_args__ = {}

    # column definitions
    guid = Column('guid', VARCHAR(length=32), primary_key=True, nullable=False, default=lambda: uuid.uuid4().hex)
    name = Column('name', VARCHAR(length=50), nullable=False)
    refcount = Column('refcount', BIGINT(), nullable=False)
    invisible = Column('invisible', INTEGER(), nullable=False)
    parent_guid = Column('parent', VARCHAR(length=32), ForeignKey('taxtables.guid'))

    # relation definitions
    entries = relation('TaxtableEntry',
                       back_populates='taxtable',
                       cascade='all, delete-orphan',
                       collection_class=CallableList,
                       )
    children = relation('Taxtable',
                        back_populates='parent',
                        cascade='all, delete-orphan',
                        collection_class=CallableList,
                        )
    parent = relation('Taxtable',
                      back_populates='children',
                      remote_side=guid,
                      )

    def __init__(self, name, entries=None):
        self.name = name
        self.refcount = 0
        self.invisible = 0
        if entries is not None:
            self.entries[:] = entries

    def __str__(self):
        if self.entries:
            return "TaxTable<{}:{}>".format(self.name, [te.__str__() for te in self.entries])
        else:
            return "TaxTable<{}>".format(self.name)


class TaxtableEntry(DeclarativeBase):
    __tablename__ = 'taxtable_entries'

    __table_args__ = {'sqlite_autoincrement': True}

    # column definitions
    id = Column('id', INTEGER(), primary_key=True, nullable=False, autoincrement=True)
    taxtable_guid = Column('taxtable', VARCHAR(length=32),
                           ForeignKey('taxtables.guid'), nullable=False)
    account_guid = Column('account', VARCHAR(length=32), ForeignKey('accounts.guid'), nullable=False)
    _amount_num = Column('amount_num', BIGINT(), nullable=False)
    _amount_denom = Column('amount_denom', BIGINT(), nullable=False)
    amount = hybrid_property_gncnumeric(_amount_num, _amount_denom)
    type = Column('type', ChoiceType({1: "value", 2: "percentage"}), nullable=False)

    # relation definitions
    taxtable = relation('Taxtable', back_populates='entries')
    account = relation('Account')

    def __init__(self, type, amount, account, taxtable=None):
        self.type = type
        self.amount = amount
        self.account = account
        if taxtable:
            self.taxtable = taxtable

    def __str__(self):
        return "TaxEntry<{} {} in {}>".format(self.amount, self.type, self.account.name)
